"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ دِگَربیان است.

[پیکرهٔ
دگربیان](https://www.peykaregan.ir/dataset/%D9%BE%DB%8C%DA%A9%D8%B1%D9%87-
%D8%AF%DA%AF%D8%B1%D8%A8%DB%8C%D8%A7%D9%86) حاوی
۱۵۲۳ نمونه است که به عنوان نمونه‌های دگربیان نشانه‌گذاری شده‌اند. جملات و
عبارات
دگربیان، بیانی متفاوت از مفهومی یکسان هستند. داده‌های این پیکره از خبرگزاری‌ها
جمع‌آورده شده و در سه دسته‌بندی «دگربیان»، «تقریباً دگربیان» و «نامرتبط» ارائه
می‌شوند. این داده‌ها با استفاده از همکاری جمعی در پیام‌رسان تلگرام نشانه‌گذاری
شده است.

"""


import os
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Tuple
from xml.dom import minidom


class DegarbayanReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ دگربیان است.

    Args:
        root: مسیر فولدر حاوی فایل‌های پیکره
        corpus_file: فایل اطلاعات پیکره.
            در صورتی که بخواهید از حالت استاندارد پیکره استفاده کنید نیازی به تغییرِ این فایل نیست.
        judge_type: این پارامتر دارای دو مقدار `three_class` و `two_class` است.
            در حالت `three_class` جملات سه برچسب می‌خورند: ۱. `Paraphrase`(دگربیان)
            ۲. `SemiParaphrase`(تقریباً دگربیان) ۳. `NotParaphrase`(غیر دگربیان). در حالت
            `two_class` حالت دوم یعنی `SemiParaphrase` هم برچسب `Paraphrase` می‌خورَد.

    """

    def __init__(
        self: "DegarbayanReader",
        root: str,
        corpus_file: str = "corpus_pair.xml",
        judge_type: str = "three_class",
    ) -> None:
        self._root = root
        self._corpus_file = corpus_file
        self._judge_type = judge_type
        if judge_type != "three_class" and judge_type != "two_class":
            self._judge_type = "three_class"

    def docs(self: "DegarbayanReader") -> Iterator[Dict[str, Any]]:
        """اسناد موجود در پیکره را برمی‌گرداند.

        Yields:
            سند بعدی.

        """

        def judge_number_to_text(judge: str) -> str:
            if judge == "1" or (self._judge_type == "two_class" and judge == "0"):
                return "Paraphrase"

            if judge == "0":
                return "SemiParaphrase"

            return "NotParaphrase"

        filename = os.path.join(self._root, self._corpus_file) # noqa: PTH118
        if os.path.exists(filename): # noqa: PTH110
            try:
                elements = minidom.parse(filename)
                for element in elements.getElementsByTagName("Pair"):
                    pair = {
                        "id": (
                            element.getElementsByTagName("PairId")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "news_source1": (
                            element.getElementsByTagName("NewsSource1")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "news_source2": (
                            element.getElementsByTagName("NewsSource2")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "news_id1": (
                            element.getElementsByTagName("NewsId1")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "news_id2": (
                            element.getElementsByTagName("NewsId2")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "sentence1": (
                            element.getElementsByTagName("Sentence1")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "sentence2": (
                            element.getElementsByTagName("Sentence2")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "method_type": (
                            element.getElementsByTagName("MethodType")[0]
                            .childNodes[0]
                            .data.strip()
                        ),
                        "judge": judge_number_to_text(
                            element.getElementsByTagName("judge")[0]
                            .childNodes[0]
                            .data.strip(),
                        ),
                    }
                    yield pair

            except Exception as e:
                print("error in reading", filename, e, file=sys.stderr)
        else:
            print("error in reading file", filename, e, file=sys.stderr)  # noqa: F821
            msg = "error in reading file"
            raise FileNotFoundError(msg, filename)

    def pairs(self: "DegarbayanReader") -> Iterator[Tuple[str, str, str]]:
        """متن‌های دگربیان را در قالب یک `(متن اصلی، شکل دگربیان، برچسب)` برمی‌گرداند.

        Examples:
            >>> degarbayan = DegarbayanReader(root='degarbayan')
            >>> next(degarbayan.pairs())
            ('24 نفر نهایی تیم ملی بدون تغییری خاص معرفی شد', 'کی روش 24 بازیکن را به تیم ملی فوتبال دعوت کرد', 'Paraphrase')

        Yields:
            `متن دگربیان بعدی در قالب یک `(متن اصلی، شکل دگربیان، برچسب).

        """
        for pair in self.docs():
            yield pair["sentence1"], pair["sentence2"], pair["judge"]



"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ تی‌نیوز است."""


import os
import re
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from xml.dom import minidom


class TNewsReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ تی‌نیوز است.

    Args:
        root: مسیر فولدر حاوی فایل‌های پیکره.

    """

    def __init__(self: "TNewsReader", root: str) -> None:
        self._root = root
        self.cleaner = re.compile(r"<[^<>]+>")

    def docs(self: "TNewsReader") -> Iterator[Dict[str, str]]:
        """خبرها را در قالب یک `iterator` برمی‌گرداند.

        هر خبر، شی‌ای متشکل از چند پارامتر است:

        - شناسه (id)،
        - عنوان (title)،
        - پیش از عنوان (pre-title)،
        - پس از عنوان (post-title)،
        - متن (text)،
        - خلاصه (brief)،
        - آدرس (url)،
        - موضوع (category)،
        - تاریخ و زمان انتشار (datetime).

        Examples:
            >>> tnews = TNewsReader(root='tnews')
            >>> next(tnews.docs())['id']
            '14092303482300013653'

        Yields:
            (Dict): خبر بعدی.

        """

        def get_text(element: str) -> str:
            raw_html = element.childNodes[0].data if element.childNodes else ""
            return re.sub(self.cleaner, "", raw_html)

        for root, _dirs, files in os.walk(self._root):
            for name in sorted(files):
                try:
                    path = Path(root) / name
                    content = path.read_text(encoding="utf8")

                    # fix xml formating issue
                    content = (
                        re.sub(
                            r"[\x1B\b\x1A]",
                            "",
                            content,
                        ).replace(
                            "</TNews>",
                            "",
                        )
                        + "</TNews>"
                    )

                    elements = minidom.parseString(content)
                    for element in elements.getElementsByTagName("NEWS"):
                        doc = {}
                        doc["id"] = get_text(element.getElementsByTagName("NEWSID")[0])
                        doc["url"] = get_text(element.getElementsByTagName("URL")[0])
                        doc["datetime"] = get_text(
                            element.getElementsByTagName("UTCDATE")[0],
                        )
                        doc["category"] = get_text(
                            element.getElementsByTagName("CATEGORY")[0],
                        )
                        doc["pre-title"] = get_text(
                            element.getElementsByTagName("PRETITLE")[0],
                        )
                        doc["title"] = get_text(
                            element.getElementsByTagName("TITLE")[0],
                        )
                        doc["post-title"] = get_text(
                            element.getElementsByTagName("POSTTITLE")[0],
                        )
                        doc["brief"] = get_text(
                            element.getElementsByTagName("BRIEF")[0],
                        )
                        doc["text"] = get_text(
                            element.getElementsByTagName("DESCRIPTION")[0],
                        )
                        yield doc

                except Exception as e:
                    print("error in reading", name, e, file=sys.stderr)

    def texts(self: "TNewsReader") -> Iterator[str]:
        """فقط متن خبرها را برمی‌گرداند.

        این تابع صرفاً برای راحتی بیشتر تهیه شده وگرنه با همان تابع
        ‍[docs()][hazm.corpus_readers.tnews_reader.TNewsReader.docs] و دریافت مقدار پراپرتی
        `text` نیز می‌توانید همین کار را انجام دهید.

        Examples:
            >>> tnews = TNewsReader(root='tnews')
            >>> next(tnews.texts()).startswith('به گزارش ”  شبکه اطلاع رسانی اینترنتی بوتیا  ” به نقل از ارگ نیوز')
            True

        Yields:
            متن خبر بعدی.

        """
        for doc in self.docs():
            yield doc["text"]



"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ تری‌بانک است.

پیکرهٔ تری‌بانک حاوی هزاران جملهٔ برچسب‌خورده با اطلاعات نحوی و ساخت‌واژی است.

"""


import os
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple
from xml.dom import minidom
from xml.dom.minidom import Node

from nltk.tree import Tree

from hazm import WordTokenizer


def coarse_pos_e(tags: List[str]) -> List[str]:
    """برچسب‌های ریز را به برچسب‌های درشت (coarse-grained pos tags) تبدیل می‌کند.

    Examples:
        >>> coarse_pos_e(['Nasp---', 'pers', 'prop'])
        'N'

    Args:
        tags: لیست برچسب‌های ریز.

    Returns:
        لیست برچسب‌های درشت.

    """
    mapping = {
        "N": "N",
        "V": "V",
        "A": "AJ",
        "D": "ADV",
        "Z": "PRO",
        "T": "DET",
        "E": "P",
        "P": "POSTP",
        "U": "NUM",
        "J": "CONJ",
        "O": "PUNC",
        "R": "RES",
        "L": "CL",
        "I": "INT",
    }
    try:
        if tags[0][0] == "C":
            if "pronominal" in tags:
                tags[0] = "Z"
            elif "verb" in tags:
                tags[0] = "V"
            elif "prep" in tags:
                tags[0] = "E"
            elif "adv" in tags:
                tags[0] = "D"
            elif "det" in tags:
                tags[0] = "T"
        return mapping[tags[0][0]] + ("e" if "ezafe" in tags else "")
    except Exception:
        return ""


class TreebankReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ تری‌بانک است.

    Args:
        root: مسیر فولدر حاوی فایل‌های پیکره
        pos_map: دیکشنری مبدل برچسب‌های ریز به درشت.
        join_clitics: اگر `True‍` باشد واژه‌بست‌ها را به کلمهٔ مادر می‌چسباند.
        join_verb_parts: اگر `True` باشد افعال چندبخشی را با _ به هم می‌چسباند.

    """

    def __init__(
        self: "TreebankReader",
        root: str,
        pos_map: str = coarse_pos_e,
        join_clitics: bool = False,
        join_verb_parts: bool = False,
    ) -> None:
        self._root = root
        self._pos_map = pos_map if pos_map else lambda tags: ",".join(tags)
        self._join_clitics = join_clitics
        self._join_verb_parts = join_verb_parts
        self._tokenizer = WordTokenizer()

    def docs(self: "TreebankReader") -> Iterator[Any]:
        """اسناد موجود در پیکره را برمی‌گرداند.

        Yields:
            سند بعدی.

        """
        def remove_blanks(node):
            for x in node.childNodes:
                if x.nodeType == Node.TEXT_NODE:
                    if x.nodeValue:
                        x.nodeValue = x.nodeValue.strip()
                elif x.nodeType == Node.ELEMENT_NODE:
                    remove_blanks(x)

        for root, _dirs, files in os.walk(self._root):
            for name in sorted(files):
                try:
                    with Path.open(
                        Path(root) / name,
                        encoding="utf8",
                    ) as treebank_file:
                        xml = minidom.parseString(treebank_file.read().encode("utf8"))
                        remove_blanks(xml)
                        xml.normalize()
                        yield xml
                except Exception as e:
                    print("error in reading", name, e, file=sys.stderr)

    def trees(self: "TreebankReader") -> Iterator[str]:
        """ساختارهای درختی موجود در پیکره را برمی‌گرداند.

        Examples:
            >>> treebank = TreebankReader(root='treebank')
            >>> print(next(treebank.trees()))
            (S
              (VPS
                (NPC (N دنیای/Ne) (MN (N آدولف/N) (N بورن/N)))
                (VPC
                  (NPC (N دنیای/Ne) (NPA (N اتفاقات/Ne) (ADJ رویایی/AJ)))
                  (V است/V)))
              (PUNC ./PUNC))


        Yields:
            ساختار درختی بعدی.

        """

        def traverse(node: str) -> Tree:
            def extract_tags(w: str):
                pos = [w.getAttribute("lc") if w.getAttribute("lc") else None]
                if w.getAttribute("clitic") in {
                    "ezafe",
                    "pronominal",
                    "verb",
                    "prep",
                    "adv",
                    "det",
                }:
                    pos.append(w.getAttribute("clitic"))
                if w.getAttribute("ne_sort"):
                    pos.append(w.getAttribute("ne_sort"))
                if w.getAttribute("n_type"):
                    pos.append(w.getAttribute("n_type"))
                if w.getAttribute("ya_type"):
                    pos.append(w.getAttribute("ya_type"))
                if w.getAttribute("ke_type"):
                    pos.append(w.getAttribute("ke_type"))
                if w.getAttribute("type"):
                    pos.append(w.getAttribute("type"))
                if w.getAttribute("kind"):
                    pos.append(w.getAttribute("kind"))
                return pos

            def clitic_join(tree: Tree, clitic: Dict):
                if type(tree[-1]) == Tree:
                    return clitic_join(tree[-1], clitic)

                if clitic[0][0][0] == "ا":
                    clitic[0] = ("‌" + clitic[0][0], clitic[0][1])
                tree[-1] = (tree[-1][0] + clitic[0][0], clitic[0][1])
                tree.set_label("CLITICS")
                return None

            if not len(node.childNodes):
                return None
            first = node.childNodes[0]
            if first.tagName == "w":
                pos = extract_tags(first)
                return Tree(
                    node.tagName,
                    [
                        (
                            first.childNodes[0].data.replace("می ", "می‌"),
                            self._pos_map(pos),
                        ),
                    ],
                )
            childs = node.childNodes[2:] if node.tagName == "S" else node.childNodes
            for child in childs:
                if not len(child.childNodes):
                    childs.remove(child)
            tree = Tree(node.tagName, map(traverse, childs))
            if (
                self._join_clitics
                and len(tree) > 1
                and type(tree[1]) == Tree
                and tree[1].label() == "CLITIC"
                and tree[1][0][1] not in {"P", "V"}
            ):
                clitic = tree[-1]
                tree = Tree(tree.label(), list(tree[0]))
                clitic_join(tree, clitic)
            if (
                self._join_verb_parts
                and len(tree) > 1
                and type(tree[1]) == Tree
                and type(tree[0]) == Tree
                and tree[0].label() == "AUX"
                and tree[0][0][0] in self._tokenizer.before_verbs
            ):
                tree[1][0] = (tree[0][0][0] + " " + tree[1][0][0], tree[1][0][1])
                tree.remove(tree[0])
            if (
                self._join_verb_parts
                and len(tree.leaves()) > 1
                and tree.leaves()[-1][0] in self._tokenizer.after_verbs
                and tree.leaves()[-2][0] in self._tokenizer.verbe
            ):
                tree[1][0] = (
                    tree[0].leaves()[-1][0] + " " + tree[1][0][0],
                    tree[1][0][1],
                )
                path = tree.leaf_treeposition(len(tree.leaves()) - 2)
                removingtree = tree
                while len(path) > 2:
                    removingtree = removingtree[path[0]]
                    path = path[1:]
                removingtree.remove(Tree(tree.pos()[-2][1], [tree.pos()[-2][0]]))
            if (
                self._join_verb_parts
                and len(tree.leaves()) > 1
                and tree.leaves()[-1][0] in self._tokenizer.after_verbs
                and tree.leaves()[-2][0] in self._tokenizer.verbe
            ):
                tree[1][0] = (
                    tree[0].leaves()[-1][0] + " " + tree[1][0][0],
                    tree[1][0][1],
                )
                path = tree.leaf_treeposition(len(tree.leaves()) - 2)
                removingtree = tree
                while len(path) > 2:
                    removingtree = removingtree[path[0]]
                    path = path[1:]
                removingtree.remove(Tree(tree.pos()[-2][1], [tree.pos()[-2][0]]))
            return tree

        for doc in self.docs():
            for s in doc.getElementsByTagName("S"):
                traverse(s)
                yield traverse(s)

    def sents(self: "TreebankReader") -> Iterator[List[Tuple[str, str]]]:
        """جملات را به شکل مجموعه‌ای از `(توکن،برچسب)`ها برمی‌گرداند.

        Examples:
            >>> treebank = TreebankReader(root='treebank')
            >>> next(treebank.sents())
            [('دنیای', 'Ne'), ('آدولف', 'N'), ('بورن', 'N'), ('دنیای', 'Ne'), ('اتفاقات', 'Ne'), ('رویایی', 'AJ'), ('است', 'V'), ('.', 'PUNC')]

        Yields:
            جملهٔ بعدی.

        """
        for tree in self.trees():
            yield tree.leaves()

    def chunked_trees(self: "TreebankReader") -> Iterator[str]:
        """ساختار درختی را به شکل تقطیع شده برمی‌گرداند.

        Examples:
            >>> from hazm.chunker import tree2brackets
            >>> treebank = TreebankReader(root='treebank')
            >>> tree2brackets(next(treebank.chunked_trees()))
            '[دنیای آدولف بورن NP] [دنیای اتفاقات رویایی NP] [است VP] .'

        Yields:
            درخت تقطیع شدهٔ بعدی.

        """

        def collapse(node, label):
            return Tree(label, [Tree(pos[1], [pos[0]]) for pos in node.pos()])

        def traverse(node, parent, chunks):
            label = node.label()

            if label.count("-nid") > 0:
                label = label.replace("-nid", "")
            if label.count("-nid") > 0:
                label = label.replace("-nid", "")
            if label.count("-DiscA") > 0:
                label = label.replace("-DiscA", "")

            if label in {"CLITIC", "CLITICS"}:
                if node[0][1] == "V":
                    label = "V"
                elif node[0][1] == "P":
                    label = "PREP"
                elif node[0][1] == "DET":
                    label = "DET"
                elif node[0][1] == "ADV":
                    label = "ADV"
                elif node[0][1] == "PRO":
                    label = "PRON"

            if label in {"CONJ", "PUNC"} and len(node) == 1:
                chunks.append(node)
                return

            if label == "PPC" and len(node) == 1:
                chunks.append(Tree("PP", [node[0]]))
                return

            if label == "PREP":
                chunks.append(Tree("PP", [node]))
                return

            if label == "PostP":
                chunks.append(Tree("POSTP", [node]))
                return

            for leaf in node.pos():
                if leaf[1] in {"PUNC", "CONJ", "PREP", "PostP"}:
                    for i in range(len(node)):
                        traverse(node[i], node, chunks)
                    return

            if label == "NPA" and parent.label() in {"CPC", "PPC"}:
                chunks.append(collapse(node, "NP"))
                return

            if label == "NPA" and len(node) >= 1 and node[0].label() == "ADV":
                chunks.append(collapse(node, "NP"))
                return

            if label in {
                "NPC",
                "N",
                "INFV",
                "DPA",
                "CLASS",
                "DPC",
                "DEM",
                "INTJ",
                "MN",
                "PRON",
                "DET",
                "NUM",
                "RES",
            }:
                chunks.append(collapse(node, "NP"))
                return

            if (
                label == "NPA"
                and len(node) >= 2
                and (
                    node[0].label() == "ADJ"
                    and node[1].label() == "NPC"
                    or node[0].label() in {"N", "PRON"}
                    and node[1].label() in {"ADJ", "ADJPA", "N"}
                    or node[0].label() == "NUM"
                    and node[1].label() in {"N", "NPC", "MN", "NUM"}
                    or node[0].label() in {"N", "NPC", "MN"}
                    and node[1].label() == "NUM"
                    or node[0].label() == "NPC"
                    and node[1].label() == "ADJ"
                    or node[0].label() == "NPA"
                    and node[1].label() != "NPC"
                    or node[1].label() == "NPA"
                    and node[0].label() != "NPC"
                )
            ):
                chunks.append(collapse(node, "NP"))
                return

            if label == "DPC" and len(node) >= 2:
                chunkable = True
                for leaf in node[1].pos():
                    if leaf[1] in {"PUNC", "CONJ", "PREP", "PostP"}:
                        chunkable = False
                if node[1].label() in {"N", "NPA", "NPC"} and chunkable:
                    chunks.append(collapse(node, "NP"))
                    return

            if label == "DPA" and len(node) >= 2 and node[1].label() == "ADV":
                chunks.append(collapse(node, "ADVP"))
                return

            if label in {"MV", "V", "AUX", "PPARV"}:
                chunks.append(Tree("VP", [node]))
                return

            if label in {"ADJ", "ADJPC", "MADJ", "ADVPA"}:
                chunks.append(Tree("ADJP", [node]))
                return

            if label in {"ADV", "MADV", "ADVPC"}:
                chunks.append(Tree("ADVP", [node]))
                return

            if type(node[0]) != Tree:
                chunks.append(node)
                return

            for i in range(len(node)):
                traverse(node[i], node, chunks)

        for tree in self.trees():
            chunks = []
            traverse(tree, None, chunks)
            for i in range(len(chunks)):
                if chunks[i].label() in {"PUNC", "CONJ"}:
                    chunks[i] = chunks[i][0]
                else:
                    chunks[i] = Tree(chunks[i].label(), chunks[i].leaves())
            yield Tree("S", chunks)



"""این ماژول شامل کلاس‌ها و توابعی برای شناساییِ وابستگی‌های دستوری متن است.

**میزان دقت
این ماژول در نسخهٔ حاضر ۹۷.۱ درصد [^1] است.**
[^1]:
این عدد با انتشار هر نسخه بروزرسانی می‌شود
"""


import os
import tempfile
from pathlib import Path
from typing import List
from typing import Tuple
from typing import Type

from nltk.parse import DependencyGraph
from nltk.parse.api import ParserI
from nltk.parse.malt import MaltParser


class MaltParser(MaltParser):
    """این کلاس شامل توابعی برای شناسایی وابستگی‌های دستوری است.

    Args:
        tagger: نام تابع `POS Tagger`.
        lemmatizer: نام کلاس ریشه‌یاب.
        working_dir: محل ذخیره‌سازی `maltparser‍`.
        model_file: آدرس مدلِ از پیش آموزش دیده با پسوند `mco`.

    """

    def __init__(
        self: "MaltParser",
        tagger: str,
        lemmatizer: str,
        working_dir: str = "dependency_parser",
        model_file: str = "langModel.mco",  # Don't rename this file
    ) -> None:
        self.tagger = tagger
        self.working_dir = working_dir
        self.mco = model_file
        self._malt_bin = os.path.join(working_dir, "malt.jar") # noqa: PTH118
        self.lemmatize = (
            lemmatizer.lemmatize if lemmatizer else lambda w, t: "_" # noqa: ARG005
        )

    def parse_sents(self: "MaltParser", sentences: str, verbose: bool = False) -> str:
        """گراف وابستگی را برمی‌گرداند.

        Args:
            sentences: جملاتی که باید گراف وابستگی آن‌ها استخراج شود.
            verbose: اگر `True` باشد وابستگی‌های بیشتری را برمی‌گرداند.

        Returns:
            گراف وابستگی.

        """
        tagged_sentences = self.tagger.tag_sents(sentences)
        return self.parse_tagged_sents(tagged_sentences, verbose)

    def parse_tagged_sents(
        self: "MaltParser",
        sentences: List[List[Tuple[str, str]]],
        verbose: bool = False,
    ) -> str:
        """گراف وابستگی‌ها را برای جملات ورودی برمی‌گرداند.

        Args:
            sentences: جملاتی که باید گراف وابستگی‌های آن استخراج شود.
            verbose: اگر `True` باشد وابستگی‌های بیشتری را برمی‌گرداند..

        Returns:
            گراف وابستگی جملات.

        Raises:
            Exception: در صورت بروز خطا یک اکسپشن عمومی صادر می‌شود.

        """
        input_file = tempfile.NamedTemporaryFile(
            prefix="malt_input.conll",
            dir=self.working_dir,
            delete=False,
        )
        output_file = tempfile.NamedTemporaryFile(
            prefix="malt_output.conll",
            dir=self.working_dir,
            delete=False,
        )

        try:
            for sentence in sentences:
                for i, (word, tag) in enumerate(sentence, start=1):
                    word = word.strip()
                    if not word:
                        word = "_"
                    input_file.write(
                        (
                            "\t".join(
                                [
                                    str(i),
                                    word.replace(" ", "_"),
                                    self.lemmatize(word, tag).replace(" ", "_"),
                                    tag,
                                    tag,
                                    "_",
                                    "0",
                                    "ROOT",
                                    "_",
                                    "_",
                                    "\n",
                                ],
                            )
                        ).encode("utf8"),
                    )
                input_file.write(b"\n\n")
            input_file.close()

            cmd = [
                "java",
                "-jar",
                self._malt_bin,
                "-w",
                self.working_dir,
                "-c",
                self.mco,
                "-i",
                input_file.name,
                "-o",
                output_file.name,
                "-m",
                "parse",
            ]
            if self._execute(cmd, verbose) != 0:
                raise Exception("MaltParser parsing failed: %s" % " ".join(cmd))

            return (
                DependencyGraph(item)
                for item in open(output_file.name, encoding="utf8").read().split("\n\n") # noqa: SIM115, PTH123
                if item.strip()
            )

        finally:
            input_file.close()
            os.remove(input_file.name) # noqa: PTH107
            output_file.close()
            os.remove(output_file.name) # noqa: PTH107


class TurboParser(ParserI):
    """interfaces [TurboParser](http://www.ark.cs.cmu.edu/TurboParser/) which you must
    manually install.

    """

    def __init__(self: "TurboParser", tagger, lemmatizer: str, model_file: str) -> None:
        self.tagger = tagger
        self.lemmatize = (
            lemmatizer.lemmatize if lemmatizer else lambda w, t: "_" # noqa: ARG005
        )

        import turboparser

        self._pturboparser = turboparser.PTurboParser()
        self.interface = self._pturboparser.create_parser()
        self.interface.load_parser_model(model_file)

    def parse_sents(
        self: "TurboParser",
        sentences: List[List[Tuple[str, str]]],
    ) -> Type[DependencyGraph]:
        """parse_sents."""
        tagged_sentences = self.tagger.tag_sents(sentences)
        return self.tagged_parse_sents(tagged_sentences)

    def tagged_parse_sents(
        self: "TurboParser",
        sentences: List[List[Tuple[str, str]]],
    ) -> Type[DependencyGraph]:
        """tagged_parse_sents."""
        input_file = tempfile.NamedTemporaryFile(
            prefix="turbo_input.conll",
            dir="dependency_parser",
            delete=False,
        )
        output_file = tempfile.NamedTemporaryFile(
            prefix="turbo_output.conll",
            dir="dependency_parser",
            delete=False,
        )

        try:
            for sentence in sentences:
                for i, (word, tag) in enumerate(sentence, start=1):
                    word = word.strip()
                    if not word:
                        word = "_"
                    input_file.write(
                        (
                            "\t".join(
                                [
                                    str(i),
                                    word.replace(" ", "_"),
                                    self.lemmatize(word, tag).replace(" ", "_"),
                                    tag,
                                    tag,
                                    "_",
                                    "0",
                                    "ROOT",
                                    "_",
                                    "_",
                                    "\n",
                                ],
                            )
                        ).encode("utf8"),
                    )
                input_file.write(b"\n")
            input_file.close()

            self.interface.parse(input_file.name, output_file.name)

            return (
                DependencyGraph(item, cell_extractor=lambda cells: cells[1:8])
                for item in open(output_file.name, encoding="utf8").read().split("\n\n") # noqa: SIM115, PTH123
                if item.strip()
            )

        finally:
            input_file.close()
            os.remove(input_file.name) # noqa: PTH107
            output_file.close()
            os.remove(output_file.name) # noqa: PTH107


class DependencyParser(MaltParser):
    """این کلاس شامل توابعی برای شناسایی وابستگی‌های دستوری است.

    این کلاس تمام توابع خود را از کلاس
    [MaltParser][hazm.dependency_parser.MaltParser] به ارث می‌برد.

    Examples:
        >>> from hazm import POSTagger, Lemmatizer, DependencyParser
        >>> parser = DependencyParser(tagger=POSTagger(model='pos_tagger.model'), lemmatizer=Lemmatizer())
        >>> parser.parse(['من', 'به', 'مدرسه', 'رفته بودم', '.']).tree().pprint()
        (من (به (مدرسه (رفته_بودم .))))
    """



"""این ماژول شامل کلاس‌ها و توابعی برای استخراج جملاتِ متن است.

برای استخراج کلمات از تابع [WordTokenizer()][hazm.WordTokenizer] استفاده کنید.

"""


import re
from typing import List

from nltk.tokenize.api import TokenizerI


class SentenceTokenizer(TokenizerI):
    """این کلاس شامل توابعی برای استخراج جملاتِ متن است."""

    def __init__(self: "SentenceTokenizer") -> None:
        self.pattern = re.compile(r"([!.?⸮؟]+)[ \n]+")

    def tokenize(self: "SentenceTokenizer", text: str) -> List[str]:
        """متن ورودی را به جملات سازندهٔ آن می‌شِکند.

        Examples:
            >>> tokenizer = SentenceTokenizer()
            >>> tokenizer.tokenize('جدا کردن ساده است. تقریبا البته!')
            ['جدا کردن ساده است.', 'تقریبا البته!']

        Args:
            text: متنی که باید جملات آن استخراج شود.

        Returns:
            فهرست جملات استخراج‌شده.

        """
        text = self.pattern.sub(r"\1\n\n", text)
        return [
            sentence.replace("\n", " ").strip()
            for sentence in text.split("\n\n")
            if sentence.strip()
        ]



"""این ماژول شامل کلاس‌ها و توابعی برای نرمال‌سازی متن‌های محاوره‌ای است."""


import re
from pathlib import Path
from typing import List

from hazm import NUMBERS
from hazm import Lemmatizer
from hazm import Normalizer
from hazm import SentenceTokenizer
from hazm import Stemmer
from hazm import WordTokenizer
from hazm import default_verbs
from hazm import informal_verbs
from hazm import informal_words


class InformalNormalizer(Normalizer):
    """این کلاس شامل توابعی برای نرمال‌سازی متن‌های محاوره‌ای است.

    Args:
        verb_file: فایل حاوی افعال محاوره‌ای.
        word_file: فایل حاوی کلمات محاوره‌ای.
        seperation_flag: اگر `True` باشد و در بخشی از متن به فاصله نیاز بود آن فاصله درج می‌شود.
        **kargs: پارامترهای نامدارِ اختیاری

    """

    def __init__(
        self: "InformalNormalizer",
        verb_file: str = informal_verbs,
        word_file: str = informal_words,
        seperation_flag: bool = False,
        **kargs: str,
    ) -> None:
        self.seperation_flag = seperation_flag
        self.lemmatizer = Lemmatizer()
        self.ilemmatizer = InformalLemmatizer()
        self.stemmer = Stemmer()
        super().__init__(**kargs)

        self.sent_tokenizer = SentenceTokenizer()
        self.word_tokenizer = WordTokenizer()

        with Path.open(verb_file, encoding="utf8") as vf:
            self.pastVerbs = {}
            self.presentVerbs = {}
            for f, i, _flag in [x.strip().split(" ", 2) for x in vf]:
                splitedf = f.split("#")
                self.presentVerbs.update({i: splitedf[1]})
                self.pastVerbs.update({splitedf[0]: splitedf[0]})
        with Path.open(default_verbs, encoding="utf8") as vf:
            for f, i in [x.strip().split("#", 2) for x in vf]:
                self.presentVerbs.update({i: i})
                self.pastVerbs.update({f: f})

        def informal_to_formal_conjucation(i, f, flag):
            iv = self.informal_conjugations(i)
            fv = self.lemmatizer.conjugation.get_all(f)
            res = {}
            if flag:
                for i, j in zip(iv, fv[48:]):
                    res[i] = j
                    if "‌" in i:
                        res[i.replace("‌", "")] = j
                        res[i.replace("‌", " ")] = j
                    if i.endswith("ین"):
                        res[i[:-1] + "د"] = j
            else:
                for i, j in zip(iv[8:], fv[56:]):
                    res[i] = j
                    if "‌" in i:
                        res[i.replace("‌", "")] = j
                        res[i.replace("‌", " ")] = j
                    if i.endswith("ین"):
                        res[i[:-1] + "د"] = j

            return res

        with Path.open(verb_file, encoding="utf8") as vf:
            self.iverb_map = {}
            for f, i, flag in [x.strip().split(" ", 2) for x in vf]:
                self.iverb_map.update(informal_to_formal_conjucation(i, f, flag))

        with Path.open(word_file, encoding="utf8") as wf:
            self.iword_map = dict([x.strip().split(" ", 1) for x in wf])

        self.words = set()
        if self.seperation_flag:
            self.words.update(list(self.iword_map.keys()))
            self.words.update(list(self.iword_map.values()))
            self.words.update(list(self.iverb_map.keys()))
            self.words.update(list(self.iverb_map.values()))
            self.words.update(self.lemmatizer.words)
            self.words.update(list(self.lemmatizer.verbs.keys()))
            self.words.update(list(self.lemmatizer.verbs.values()))

    def split_token_words(self: "InformalNormalizer", token: str) -> str:
        """هرجایی در متن فاصله نیاز بود قرار می‌دهد.

        متأسفانه در برخی از متن‌ها، به بهانهٔ صرفه‌جویی در زمان یا از سرِ تنبلی،
        فاصله‌گذاری‌ها درست رعایت نمی‌شود. مثلاً جملهٔ «تو را دوست دارم.» به این
        شکل نوشته می‌شود: «تورادوست دارم.» این تابع فواصل ضروری را در متن
        ایجاد می‌کند و آن را به شکل صحیح برمی‌گرداند.

        Args:
            token: توکنی که باید فاصله‌گذاری شود.

        Returns:
            توکنی با فاصله‌گذاری صحیح.

        """

        def shekan(token):
            res = [""]
            for i in token:
                res[-1] += i
                if i in {"ا", "د", "ذ", "ر", "ز", "ژ", "و", *list(NUMBERS)}:
                    res.append("")
            while "" in res:
                res.remove("")
            return res

        def perm(lst):
            if len(lst) > 1:
                up = perm(lst[1:])
            else:
                return [lst]
            res = []
            for i in up:
                res.append([lst[0], *i])
                res.append([lst[0] + i[0]] + i[1:])
            res.sort(key=len)
            return res

        token = re.sub(r"(.)\1{2,}", r"\1", token)
        ps = perm(shekan(token))
        for c in ps:
            if {self.ilemmatizer.lemmatize(x) for x in c}.issubset(self.words):
                return " ".join(c)
        return token

    def normalized_word(self: "InformalNormalizer", word: str) -> List[str]:
        """اشکال مختلف نرمالایزشدهٔ کلمه را برمی‌گرداند.

        Examples:
            >>> normalizer = InformalNormalizer()
            >>> normalizer.normalized_word('می‌رم')
            ['می‌روم', 'می‌رم']

        Args:
            word: کلمه‌ای که باید نرمال‌سازی شود.

        Returns:
            اشکال نرمالایزشدهٔ کلمه.

        """
        # >>> normalizer = InformalNormalizer(seperation_flag=True)
        # >>> normalizer.normalized_word('صداوسیماجمهوری')

        def analyze_word(word):
            end_words_list = [
                "هاست",
                "هایی",
                "هایم",
                "ترین",
                "ایی",
                "انی",
                "شان",
                "شون",
                "است",
                "تان",
                "تون",
                "مان",
                "مون",
                "هام",
                "هاش",
                "های",
                "طور",
                "ها",
                "تر",
                "ئی",
                "یی",
                "یم",
                "ام",
                "ای",
                "ان",
                "هم",
                "رو",
                "یت",
                "ه",
                "ی",
                "ش",
                "و",
                "ا",
                "ت",
                "م",
            ]

            return_list = []

            collection_of_word_and_suffix = []

            found_early = False

            mid_word_condidate = []

            if word.endswith(("\u200c", "\u200e")):
                word = word[:-1]

            if word in self.lemmatizer.words or word in self.iword_map:
                if word in self.lemmatizer.words:
                    collection_of_word_and_suffix.append({"word": word, "suffix": []})
                if word in self.iword_map:
                    collection_of_word_and_suffix.append(
                        {"word": self.iword_map[word], "suffix": []},
                    )
                found_early = True

            if not found_early:
                for endword in end_words_list:
                    if word.endswith(endword):
                        sliceword = word[: -1 * len(endword)]
                        if (
                            sliceword in self.lemmatizer.words
                            or sliceword in self.iword_map
                        ):
                            if sliceword in self.lemmatizer.words:
                                collection_of_word_and_suffix.append(
                                    {"word": sliceword, "suffix": [endword]},
                                )
                            if sliceword in self.iword_map:
                                collection_of_word_and_suffix.append(
                                    {
                                        "word": self.iword_map[sliceword],
                                        "suffix": [endword],
                                    },
                                )
                        else:
                            mid_word_condidate.append(sliceword)
                            mid_word_condidate.append([endword])

                for endword in end_words_list:
                    for i in range(len(mid_word_condidate) - 1):
                        if i % 2 == 1:
                            continue
                        midword = mid_word_condidate[i]
                        mid_word_end_word_list = mid_word_condidate[i + 1]
                        if midword.endswith(endword):
                            sliceword = midword[: -1 * len(endword)]
                            if (
                                sliceword in self.lemmatizer.words
                                or sliceword in self.iword_map
                            ):
                                if sliceword in self.lemmatizer.words:
                                    collection_of_word_and_suffix.append(
                                        {
                                            "word": sliceword,
                                            "suffix": [
                                                endword,
                                                *mid_word_end_word_list,
                                            ],
                                        },
                                    )
                                if sliceword in self.iword_map:
                                    collection_of_word_and_suffix.append(
                                        {
                                            "word": self.iword_map[sliceword],
                                            "suffix": [
                                                endword,
                                                *mid_word_end_word_list,
                                            ],
                                        },
                                    )

            for i in range(len(collection_of_word_and_suffix)):
                new_possible_word_list = append_suffix_to_word(
                    collection_of_word_and_suffix[i],
                )
                for j in range(len(new_possible_word_list)):
                    new_possible_word = new_possible_word_list[j]
                    if new_possible_word not in return_list:
                        return_list.append(new_possible_word)

            return return_list

        def analyze_verb_word(word):
            if word in self.pastVerbs:
                word = self.pastVerbs[word]
                return [word]

            if word in self.iword_map:
                return []

            if word in self.lemmatizer.words:
                if word[-1] == "ن":
                    None  # noqa: B018
                else:
                    return []

            return_list = []

            collection_of_verb_list = []

            end_verb_list = [
                "یم",
                "دم",
                "دیم",
                "ید",
                "دی",
                "دید",
                "ند",
                "دن",
                "دند",
                "ین",
                "دین",
                "ست",
                "ستم",
                "ستی",
                "ستیم",
                "ستید",
                "ستند",
                "ستن",
                "م",
                "ی",
                "ه",
                "د",
                "ن",
            ]

            for endverb in end_verb_list:
                if word.endswith(endverb):
                    if endverb == "ین":
                        collection_of_verb_list.append(
                            {"word": word[:-2], "suffix": "ید"},
                        )
                    elif endverb == "ستن":
                        collection_of_verb_list.append(
                            {"word": word[:-3], "suffix": "ستند"},
                        )
                    elif endverb == "ن":
                        collection_of_verb_list.append(
                            {"word": word[:-1], "suffix": "ن"},
                        )
                        collection_of_verb_list.append(
                            {"word": word[:-1], "suffix": "ند"},
                        )
                    elif endverb == "ه":
                        if len(word) > 1:
                            if word[-2] != "د":
                                collection_of_verb_list.append(
                                    {"word": word[:-1], "suffix": "د"},
                                )
                            collection_of_verb_list.append(
                                {"word": word[:-1], "suffix": "ه"},
                            )
                        else:
                            collection_of_verb_list.append(
                                {"word": word[:-1], "suffix": "ه"},
                            )
                    else:
                        collection_of_verb_list.append(
                            {
                                "word": word[: -1 * len(endverb)],
                                "suffix": endverb,
                            },
                        )
            collection_of_verb_list.append({"word": word, "suffix": ""})
            collection_of_verb_list_2 = []
            for i in range(len(collection_of_verb_list)):
                mainword = collection_of_verb_list[i]["word"]
                collection_of_verb_list[i]["preffix"] = ""
                if mainword.startswith("بر"):
                    modified_word = mainword[2:]
                    new_main_word = ""
                    if modified_word.startswith("نمی"):
                        collection_of_verb_list[i]["preffix"] = "برنمی"
                        new_main_word = modified_word[3:]
                    elif modified_word.startswith("می"):
                        collection_of_verb_list[i]["preffix"] = "برمی"
                        new_main_word = modified_word[2:]
                    elif modified_word.startswith("ن"):
                        collection_of_verb_list[i]["preffix"] = "برن"
                        new_main_word = modified_word[1:]
                    elif modified_word.startswith("بی"):
                        collection_of_verb_list[i]["preffix"] = "بربی"
                        new_main_word = modified_word[2:]
                    elif modified_word.startswith("ب"):
                        collection_of_verb_list[i]["preffix"] = "برب"
                        new_main_word = modified_word[1:]
                    else:
                        collection_of_verb_list[i]["preffix"] = "بر"
                        new_main_word = modified_word
                        collection_of_verb_list_2.append(
                            {
                                "word": mainword,
                                "preffix": "",
                                "suffix": collection_of_verb_list[i]["suffix"],
                            },
                        )

                    if new_main_word:
                        collection_of_verb_list[i]["word"] = new_main_word
                elif mainword.startswith("نمی"):
                    collection_of_verb_list[i]["preffix"] = "نمی"
                    collection_of_verb_list[i]["word"] = mainword[3:]
                elif mainword.startswith("می"):
                    collection_of_verb_list[i]["preffix"] = "می"
                    collection_of_verb_list[i]["word"] = mainword[2:]
                elif mainword.startswith("ن"):
                    collection_of_verb_list[i]["preffix"] = "ن"
                    collection_of_verb_list[i]["word"] = mainword[1:]
                    collection_of_verb_list_2.append(
                        {
                            "word": mainword,
                            "preffix": "",
                            "suffix": collection_of_verb_list[i]["suffix"],
                        },
                    )

                elif mainword.startswith("بی"):
                    collection_of_verb_list[i]["preffix"] = "بی"
                    collection_of_verb_list[i]["word"] = mainword[2:]
                elif mainword.startswith("ب"):
                    collection_of_verb_list[i]["preffix"] = "ب"
                    collection_of_verb_list[i]["word"] = mainword[1:]
                    collection_of_verb_list_2.append(
                        {
                            "word": mainword,
                            "preffix": "",
                            "suffix": collection_of_verb_list[i]["suffix"],
                        },
                    )

            for i in range(len(collection_of_verb_list_2)):
                collection_of_verb_list.append(collection_of_verb_list_2[i])

            collection_of_real_verb_list = []
            for i in range(len(collection_of_verb_list)):
                mainword = collection_of_verb_list[i]["word"]
                if mainword.startswith(("\u200c", "\u200e")):
                    mainword = mainword[1:]

                mainword2 = None
                if mainword.startswith("ا"):
                    mainword2 = "آ" + mainword[1:]
                if mainword in self.pastVerbs:
                    collection_of_verb_list[i]["word"] = self.pastVerbs[mainword]
                    collection_of_real_verb_list.append(collection_of_verb_list[i])
                if mainword in self.presentVerbs:
                    collection_of_verb_list[i]["word"] = self.presentVerbs[mainword]
                    collection_of_real_verb_list.append(collection_of_verb_list[i])
                if mainword2 is not None and not (
                    collection_of_verb_list[i]["preffix"] == "بربی"
                    or collection_of_verb_list[i]["preffix"] == "بی"
                ):
                    if mainword2 in self.pastVerbs:
                        collection_of_verb_list[i]["word"] = self.pastVerbs[mainword2]
                        collection_of_real_verb_list.append(collection_of_verb_list[i])
                    if mainword2 in self.presentVerbs:
                        collection_of_verb_list[i]["word"] = self.presentVerbs[
                            mainword2
                        ]
                        collection_of_real_verb_list.append(collection_of_verb_list[i])

            for i in range(len(collection_of_real_verb_list)):
                preffix = collection_of_real_verb_list[i]["preffix"]
                suffix = collection_of_real_verb_list[i]["suffix"]
                mainword = collection_of_real_verb_list[i]["word"]
                returnword = preffix
                if preffix.endswith("می"):
                    returnword += "‌"
                returnword += mainword
                returnword += suffix
                if mainword and returnword not in return_list:
                    return_list.append(returnword)

            return return_list

        def append_suffix_to_word(one_collection_of_word_and_suffix):
            mainword = one_collection_of_word_and_suffix["word"]
            suffix_list = one_collection_of_word_and_suffix["suffix"]
            adhesive_alphabet = {
                "ب": "ب",
                "پ": "پ",
                "ت": "ت",
                "ث": "ث",
                "ج": "ج",
                "چ": "چ",
                "ح": "ح",
                "خ": "خ",
                "س": "س",
                "ش": "ش",
                "ص": "ص",
                "ض": "ض",
                "ع": "ع",
                "غ": "غ",
                "ف": "ف",
                "ق": "ق",
                "ک": "ک",
                "گ": "گ",
                "ل": "ل",
                "م": "م",
                "ن": "ن",
                "ه": "ه",
                "ی": "ی",
            }
            return_list = []
            returnword = mainword
            return_word2 = None
            return_word3 = None
            if len(suffix_list) == 0:
                return [returnword]
            if len(suffix_list) > 1:
                if suffix_list[0] == "ه" and suffix_list[1] == "ا":
                    suffix_list[0] = "ها"
                    suffix_list.remove(suffix_list[1])
                if suffix_list[0] == "ه" and suffix_list[1] == "است":
                    suffix_list[0] = "هاست"
                    suffix_list.remove(suffix_list[1])
                if suffix_list[0] == "ت" and suffix_list[1] == "ا":
                    suffix_list[0] = "تا"
                    suffix_list.remove(suffix_list[1])
            for i in range(len(suffix_list)):
                if suffix_list[i] == "شون":
                    returnword += "شان"
                elif suffix_list[i] == "تون":
                    returnword += "تان"
                elif suffix_list[i] == "مون":
                    returnword += "مان"
                elif suffix_list[i] == "هام":
                    try:
                        adhesive_alphabet[returnword[-1]]
                        returnword += "‌"
                    except:
                        None  # noqa: B018
                    returnword += "هایم"
                elif suffix_list[i] == "ها":  # noqa: SIM114
                    try:
                        adhesive_alphabet[returnword[-1]]
                        returnword += "‌"
                    except:
                        None  # noqa: B018
                    returnword += "ها"
                elif (
                    suffix_list[i] == "ا"
                    and suffix_list[len(suffix_list) - 1] == "ا"
                    and not returnword.endswith("ه")
                ):
                    try:
                        adhesive_alphabet[returnword[-1]]
                        returnword += "‌"
                    except:
                        None  # noqa: B018
                    returnword += "ها"
                elif suffix_list[i] == "و" and suffix_list[len(suffix_list) - 1] == "و":
                    return_word2 = returnword
                    return_word2 += " و"
                    returnword += " را"

                elif (
                    suffix_list[i] == "رو" and suffix_list[len(suffix_list) - 1] == "رو"
                ):
                    returnword += " را"

                elif suffix_list[i] == "ه" and suffix_list[len(suffix_list) - 1] == "ه":
                    return_word2 = returnword
                    return_word2 += "ه"
                    return_word3 = returnword
                    return_word3 += " است"
                    returnword += "ه است"
                else:
                    returnword += suffix_list[i]
            return_list.append(returnword)
            if return_word2 is not None:
                return_list.append(return_word2)
            if return_word3 is not None:
                return_list.append(return_word3)
            return return_list

        def straight_forward_result(word):
            straight_forward_dic = {
                "ب": ["به"],
                "ک": ["که"],
                "آن": ["آن"],
                "می‌آید": ["می‌آید"],
                "میاید": ["می‌آید"],
                "می‌آیم": ["می‌آیم"],
                "میایم": ["می‌آیم"],
                "نمی‌آید": ["نمی‌آید"],
                "نمیاید": ["نمی‌آید"],
                "نمی‌آیم": ["نمی‌آیم"],
                "نمیایم": ["نمی‌آیم"],
                "برمی‌آید": ["برمی‌آید"],
                "برمیاید": ["برمی‌آید"],
                "برمی‌آیم": ["برمی‌آیم"],
                "برمیایم": ["برمی‌آیم"],
                "برنمی‌آید": ["برنمی‌آید"],
                "برنمیاید": ["برنمی‌آید"],
                "برنمی‌آیم": ["برنمی‌آیم"],
                "برنمیایم": ["برنمی‌آیم"],
                "منظوره": ["منظوره"],
                "بدن": ["بدن"],
                "میا": ["میا"],
                "نیس": ["نیست"],
                "فک": ["فکر"],
                "برام": ["برایم"],
                "آ": ["آ"],
                "آی": ["آی"],
                "این": ["این"],
                "است": ["است"],
                "ان": ["ان"],
                "اند": ["اند"],
                "میان": ["میان"],
                "گردن": ["گردن"],
                "اینهمه": ["اینهمه"],
                "آنهمه": ["آنهمه"],
                "الیه": ["الیه"],
                "غرغره": ["غرغره"],
                "لیله": ["لیله"],
                "بزرگانه": ["بزرگانه"],
                "پرستانه": ["پرستانه"],
                "ام": ["ام"],
                "بادی": ["بادی"],
                "نان": ["نان"],
                "باورم": ["باورم"],
                "اوه": ["اوه"],
                "چقد": ["چقدر"],
                "چو": ["چون"],
                "هس": ["هست"],
                "اومدند": ["آمدند"],
                "ش": ["اش"],
                "بش": ["بهش"],
                "ازت": ["از تو"],
                "رو": ["را", "رو"],
                "پایتون": ["پایتون"],
                "اردن": ["اردن"],
                "دست": ["دست"],
                "دستی": ["دستی"],
                "دستم": ["دستم"],
                "دین": ["دین"],
                "شین": ["شین"],
                "سراتو": ["سراتو"],
                "فالو": ["فالو"],
                "هرجا": ["هرجا"],
                "میدان": ["میدان"],
                "میدون": ["میدان"],
                "کفا": ["کفا"],
                "ویا": ["و یا"],
                "نشد": ["نشد"],
                "شو": ["شو"],
                "مشیا": ["مشیا"],
                "پلاسما": ["پلاسما"],
                "فیلیمو": ["فیلیمو"],
                "پاشو": ["پاشو"],
                "میر": ["میر"],
                "بارم": ["بار هم", "بارم"],
                "برند": ["برند"],
                "کنه": ["کند"],
                "بتونه": ["بتواند"],
                "باشه": ["باشد"],
                "بخوان": ["بخوان"],
                "بدم": ["بدم"],
                "برم": ["برم"],
                "بده": ["بده"],
                "نده": ["نده"],
                "شهرو": ["شهرو"],
                "شیرو": ["شیرو"],
                "نگذاشته": ["نگذاشته"],
                "نگرفته": ["نگرفته"],
                "نمیشناخته": ["نمی‌شناخته"],
                "نمی‌شناخته": ["نمی‌شناخته"],
                "بشین": ["بشین"],
                "هارو": ["ها را"],
                "مارو": ["ما را"],
                "میاومد": ["می‌آمد"],
                "می‌اومد": ["می‌آمد"],
                "میخواسته": ["می‌خواسته"],
                "می‌خواسته": ["می‌خواسته"],
                "نمیخواسته": ["نمی‌خواسته"],
                "نمی‌خواسته": ["نمی‌خواسته"],
                "میتوانسته": ["می‌توانسته"],
                "می‌توانسته": ["می‌توانسته"],
                "میرفته": ["می‌رفته"],
                "می‌رفته": ["می‌رفته"],
                "نشین": ["نشین"],
                "انا": ["انا"],
                "خونی": ["خونی"],
                "خون": ["خون"],
                "یالا": ["یالا"],
                "میخواند": ["می‌خواند"],
                "می‌خواند": ["می‌خواند"],
                "نمیخواند": ["نمی‌خواند"],
                "نمی‌خواند": ["نمی‌خواند"],
                "میده": ["می‌دهد"],
                "می‌ده": ["می‌دهد"],
                "میشه": ["می‌شود"],
                "می‌شه": ["می‌شود"],
                "میشد": ["می‌شد"],
                "می‌شد": ["می‌شد"],
                "میشدم": ["می‌شدم"],
                "می‌شدم": ["می‌شدم"],
                "نمیشد": ["نمی‌شد"],
                "نمی‌شد": ["نمی‌شد"],
                "بردم": ["بردم"],
                "بره": ["بره", "برود"],
                "شم": ["بشوم"],
                "اوست": ["اوست"],
                "بیا": ["بیا"],
                "نیا": ["نیا"],
                "میاد": ["می‌آید"],
                "نشدی": ["نشدی"],
                "بخواند": ["بخواند"],
                "سیا": ["سیا"],
                "میدید": ["می‌دید"],
                "می‌دید": ["می‌دید"],
                "وا": ["وا"],
                "برگشته": ["برگشته"],
                "میخواست": ["می‌خواست"],
                "می‌خواست": ["می‌خواست"],
            }
            try:
                return straight_forward_dic[word]
            except:
                return []

        straight_forward_words = straight_forward_result(word)
        if len(straight_forward_words) > 0:
            return straight_forward_words

        verb_words_list = analyze_verb_word(word)
        if len(verb_words_list) > 0:
            return verb_words_list
        possible_words = analyze_word(word)

        mainword = word
        if mainword in possible_words:
            possible_words.remove(mainword)
            possible_words.append(mainword)
        elif len(possible_words) == 0:
            possible_words.append(mainword)

        return possible_words

    def normalize(self: "InformalNormalizer", text: str) -> List[List[List[str]]]:
        """متن محاوره‌ای را به متن فارسی معیار تبدیل می‌کند.

        Examples:
            >>> normalizer = InformalNormalizer()
            >>> normalizer.normalize('بابا یه شغل مناسب واسه بچه هام پیدا کردن که به جایی برنمیخوره !')
            [[['بابا'], ['یک'], ['شغل'], ['مناسب'], ['برای'], ['بچه'], ['هایم'], ['پیدا'], ['کردن', 'کردند'], ['که'], ['به'], ['جایی'], ['برنمی\u200cخورد', 'برنمی\u200cخوره'], ['!']]]
            >>> normalizer = InformalNormalizer()
            >>> normalizer.normalize('اجازه بدیم همسرمون در جمع خانواده‌اش احساس آزادی کنه و فکر نکنه که ما دائم هواسمون بهش هست .')
            [[['اجازه'], ['بدهیم'], ['همسرمان'], ['در'], ['جمع'], ['خانواده\u200cاش'], ['احساس'], ['آزادی'], ['کند'], ['و'], ['فکر'], ['نکند', 'نکنه'], ['که'], ['ما'], ['دائم'], ['حواسمان'], ['بهش'], ['هست'], ['.']]]

        Args:
            text: متن محاوره‌ای که باید تبدیل به متن فارسی معیار شود.

        Returns:
           متن فارسی معیار.

        """
        text = super().normalize(text)
        sents = [
            self.word_tokenizer.tokenize(sentence)
            for sentence in self.sent_tokenizer.tokenize(text)
        ]

        return [[self.normalized_word(word) for word in sent] for sent in sents]

    def informal_conjugations(self: "InformalNormalizer", verb: str) -> List[str]:
        """صورت‌های صرفی فعل را در شکل محاوره‌ای تولید می‌کند.

        Args:
            verb: فعلی که باید صرف شود.

        Returns:
            صورت‌های صرفی فعل.

        """
        ends = ["م", "ی", "", "یم", "ین", "ن"]
        present_simples = [verb + end for end in ends]
        if verb.endswith("ا"):
            present_simples[2] = verb + "د"
        else:
            present_simples[2] = verb + "ه"
        present_not_simples = ["ن" + item for item in present_simples]
        present_imperfects = ["می‌" + item for item in present_simples]
        present_not_imperfects = ["ن" + item for item in present_imperfects]
        present_subjunctives = [
            item if item.startswith("ب") else "ب" + item for item in present_simples
        ]
        present_not_subjunctives = ["ن" + item for item in present_simples]
        return (
            present_simples
            + present_not_simples
            + present_imperfects
            + present_not_imperfects
            + present_subjunctives
            + present_not_subjunctives
        )


class InformalLemmatizer(Lemmatizer):
    """این کلاس شامل توابعی برای نرمال‌سازی متن‌های محاوره‌ای است."""

    def __init__(self: "InformalNormalizer", **kargs: str) -> None:
        """__init__."""
        super().__init__(**kargs)

        temp = []
        self.words = set(self.words.keys())
        for word in self.words:
            if word.endswith("ً"):
                temp.append(word[:-1])

        self.words.update(temp)

        temp = {}
        for verb in self.verbs:
            if verb.endswith("د"):
                temp[verb[:-1] + "ن"] = self.verbs[verb]

        self.verbs.update(temp)

        with Path.open(informal_verbs, encoding="utf8") as vf:
            for f, i, _flag in [x.strip().split(" ", 2) for x in vf]:
                self.verbs.update({x: f for x in self.iconjugations(i)})

        with Path.open(informal_words, encoding="utf8") as wf:
            self.words.update([x.strip().split(" ", 1)[0] for x in wf])

    def iconjugations(self: "InformalNormalizer", verb: str):
        """iconjugations."""
        ends = ["م", "ی", "", "یم", "ین", "ن"]
        present_simples = [verb + end for end in ends]
        if verb.endswith("ا"):
            present_simples[2] = verb + "د"
        else:
            present_simples[2] = verb + "ه"
        present_not_simples = ["ن" + item for item in present_simples]
        present_imperfects = ["می‌" + item for item in present_simples]
        present_not_imperfects = ["ن" + item for item in present_imperfects]
        present_subjunctives = [
            item if item.startswith("ب") else "ب" + item for item in present_simples
        ]
        present_not_subjunctives = ["ن" + item for item in present_simples]
        return (
            present_simples
            + present_not_simples
            + present_imperfects
            + present_not_imperfects
            + present_subjunctives
            + present_not_subjunctives
        )



"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ بی‌جن‌خان است.

[پیکرهٔ
بی‌جن‌خان](https://www.peykaregan.ir/dataset/%D9%BE%DB%8C%DA%A9%D8%B1%D9%87-
%D8%A8%DB%8C%E2%80%8C%D8%AC%D9%86%E2%80%8C%D8%AE%D8%A7%D9%86) مجموعه‌ای
است از متون فارسی شامل بیش از ۲ میلیون و ۶۰۰ هزار کلمه که با ۵۵۰ نوع برچسب POS
برچسب‌گذاری شده‌اند. این پیکره که در پژوهشکدهٔ پردازش هوشمند علائم تهیه شده است
همچنین شامل بیش از ۴۳۰۰ تگ موضوعی چون سیاسی، تاریخی و ... برای متون است.

"""

import re
from pathlib import Path
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from hazm import Normalizer

from .peykare_reader import join_verb_parts

default_pos_map = {
    "ADJ": "ADJ",
    "ADJ_CMPR": "ADJ",
    "ADJ_INO": "ADJ",
    "ADJ_ORD": "ADJ",
    "ADJ_SIM": "ADJ",
    "ADJ_SUP": "ADJ",
    "ADV": "ADV",
    "ADV_EXM": "ADV",
    "ADV_I": "ADV",
    "ADV_NEGG": "ADV",
    "ADV_NI": "ADV",
    "ADV_TIME": "ADV",
    "AR": "AR",
    "CON": "CONJ",
    "DEFAULT": "DEFAULT",
    "DELM": "PUNC",
    "DET": "PREP",
    "IF": "IF",
    "INT": "INT",
    "MORP": "MORP",
    "MQUA": "MQUA",
    "MS": "MS",
    "N_PL": "N",
    "N_SING": "N",
    "NN": "NN",
    "NP": "NP",
    "OH": "OH",
    "OHH": "OHH",
    "P": "PREP",
    "PP": "PP",
    "PRO": "PR",
    "PS": "PS",
    "QUA": "QUA",
    "SPEC": "SPEC",
    "V_AUX": "V",
    "V_IMP": "V",
    "V_PA": "V",
    "V_PRE": "V",
    "V_PRS": "V",
    "V_SUB": "V",
}


class BijankhanReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ بی‌جن‌خان است.

    Args:
        bijankhan_file: مسیر فایلِ پیکره.
        joined_verb_parts: اگر `True‍` باشد افعال چندبخشی را با _ به‌هم می‌چسباند.
        pos_map: دیکشنری مبدل برچسب‌های ریز به درشت.

    """

    def __init__(
        self: "BijankhanReader",
        bijankhan_file: str,
        joined_verb_parts: bool = True,
        pos_map: Optional[str] = None,
    ) -> None:
        """Init."""
        if pos_map is None:
            pos_map = default_pos_map
        self._bijankhan_file = bijankhan_file
        self._joined_verb_parts = joined_verb_parts
        self._pos_map = pos_map
        self._normalizer = Normalizer(correct_spacing=False)

    def _sentences(self: "BijankhanReader") -> Iterator[List[Tuple[str, str]]]:
        """جملات پیکره را به شکل متن خام برمی‌گرداند.

        Yields:
            جملهٔ بعدی.

        """
        sentence = []
        with Path(self._bijankhan_file).open(encoding="utf-8") as f:
            length = 2
            for line in f:
                parts = re.split("  +", line.strip())
                if len(parts) == length:
                    word, tag = parts
                    if word not in ("#", "*"):
                        word = self._normalizer.normalize(word)
                        sentence.append((word if word else "_", tag))
                    if (
                        tag == "DELM"
                        and word in ("#", "*", ".", "؟", "!")
                        and len(sentence)
                    ):
                        yield sentence
                        sentence = []

    def sents(self: "BijankhanReader") -> Iterator[List[Tuple[str, str]]]:
        """جملات پیکره را به شکل لیستی از `(توکن،برچسب)`ها برمی‌گرداند..

        Examples:
            >>> bijankhan = BijankhanReader(bijankhan_file='bijankhan.txt')
            >>> next(bijankhan.sents())
            [('اولین', 'ADJ'), ('سیاره', 'N'), ('خارج', 'ADJ'), ('از', 'PREP'), ('منظومه', 'N'), ('شمسی', 'ADJ'), ('دیده_شد', 'V'), ('.', 'PUNC')]

        Yields:
            جملهٔ بعدی در قالب لیستی از `(توکن،برچسب)`ها.

        """

        def map_poses(item: Tuple[str, str]) -> Tuple[str, str]:
            return (item[0], self._pos_map.get(item[1], item[1]))

        for sentence in self._sentences():
            if self._joined_verb_parts:
                sentence = join_verb_parts(sentence)
            yield list(map(map_poses, sentence))



"""این ماژول شامل کلاس‌ها و توابعی برای تبدیل کلمه یا متن به برداری از اعداد است."""
import multiprocessing
import os
import warnings
from pathlib import Path
from typing import Any
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import fasttext as fstxt
import numpy as np
import smart_open
from gensim.models import Doc2Vec
from gensim.models import KeyedVectors
from gensim.models import fasttext
from gensim.models.callbacks import CallbackAny2Vec
from gensim.models.doc2vec import TaggedDocument
from gensim.scripts.glove2word2vec import glove2word2vec
from gensim.test.utils import datapath
from numpy import ndarray

from hazm import Normalizer
from hazm import word_tokenize

supported_embeddings = ["fasttext", "keyedvector", "glove"]


class WordEmbedding:
    """این کلاس شامل توابعی برای تبدیل کلمه به برداری از اعداد است.

    Args:
        model_type: نوع امبدینگ که می‌تواند یکی از مقادیر ‍`fasttext`, `keyedvector`, `glove` باشد.
        model_path: مسیر فایل امبدینگ.

    """

    def __init__(
        self: "WordEmbedding",
        model_type: str,
        model_path: Optional[str] = None,
    ) -> None:
        if model_type not in supported_embeddings:
            msg = (
                f'Model type "{model_type}" is not supported! Please choose from'
                f" {supported_embeddings}"
            )
            raise KeyError(
                msg,
            )
        self.model_type = model_type
        if model_path:
            self.load_model(model_path)

    def load_model(self: "WordEmbedding", model_path: str) -> None:
        """فایل امبدینگ را بارگزاری می‌کند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('word2vec.bin') # doctest: +ELLIPSIS
            ...

        Args:
            model_path: مسیر فایل امبدینگ.

        """
        if self.model_type == "fasttext":
            self.model = fasttext.load_facebook_model(model_path).wv
        elif self.model_type == "keyedvector":
            if model_path.endswith("bin"):
                self.model = KeyedVectors.load_word2vec_format(model_path, binary=True)
            else:
                self.model = KeyedVectors.load_word2vec_format(model_path)
        elif self.model_type == "glove":
            word2vec_addr = str(model_path) + "_word2vec_format.vec"
            if not Path.exists(word2vec_addr):
                _ = glove2word2vec(model_path, word2vec_addr)
            self.model = KeyedVectors.load_word2vec_format(word2vec_addr)
            self.model_type = "keyedvector"
        else:
            msg = (
                f"{self.model_type} not supported! Please choose from"
                f" {supported_embeddings}"
            )
            raise KeyError(
                msg,
            )

    def train(
        self: "WordEmbedding",
        dataset_path: str,
        workers: int = multiprocessing.cpu_count() - 1,  # noqa: B008
        vector_size: int = 200,
        epochs: int = 10,
        min_count: int = 5,
        fasttext_type: str = "skipgram",
        dest_path: str = "fasttext_word2vec_model.bin",
    ) -> None:
        """یک فایل امبدینگ از نوع fasttext ترین می‌کند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.train(dataset_path = 'dataset.txt', workers = 4, vector_size = 300, epochs = 30, fasttext_type = 'cbow', dest_path = 'fasttext_model') # doctest: +ELLIPSIS
            ...

        Args:
            dataset_path: مسیر فایل متنی.
            workers: تعداد هسته درگیر برای ترین مدل.
            vector_size: طول وکتور خروجی به ازای هر کلمه.
            epochs: تعداد تکرار ترین بر روی کل دیتا.
            min_count:  حداقل تعداد تکرار یک کلمه برای قرارگیری آن در مدل امبدینگ.
            fasttext_type: نوع fasttext مورد نظر برای ترین که میتواند یکی از مقادیر skipgram یا cbow را داشته باشد.
            dest_path: مسیر مورد نظر برای ذخیره فایل امبدینگ.

        """
        if self.model_type != "fasttext":
            self.model = "fasttext"
            warnings.warn(
                (
                    "this function is for training fasttext models only and"
                    f" {self.model_type} is not supported"
                ),
                stacklevel=2,
            )

        fasttext_model_types = ["cbow", "skipgram"]
        if fasttext_type not in fasttext_model_types:
            msg = (
                f'Model type "{fasttext_type}" is not supported! Please choose from'
                f" {fasttext_model_types}"
            )
            raise KeyError(
                msg,
            )

        workers = 1 if workers == 0 else workers

        print("training model...")
        model = fstxt.train_unsupervised(
            dataset_path,
            model=fasttext_type,
            dim=vector_size,
            epoch=epochs,
            thread=workers,
            min_count = min_count,
        )

        print("Model trained.")

        print("saving model...")
        model.save_model(dest_path)
        print("Model saved.")

        print("loading model...")
        self.load_model(model_path=dest_path)
        print("model loaded.")

    def __getitem__(self: "WordEmbedding", word: str) -> Type[ndarray]:
        """__getitem__."""
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)
        return self.model[word]

    def doesnt_match(self: "WordEmbedding", words: List[str]) -> str:
        """لیستی از کلمات را دریافت می‌کند و کلمهٔ نامرتبط را برمی‌گرداند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('word2vec.bin')
            >>> wordEmbedding.doesnt_match(['سلام' ,'درود' ,'خداحافظ' ,'پنجره'])
            'پنجره'
            >>> wordEmbedding.doesnt_match(['ساعت' ,'پلنگ' ,'شیر'])
            'ساعت'

        Args:
            words: لیست کلمات.

        Returns:
            کلمهٔ نامرتبط.

        """
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)
        return self.model.doesnt_match(words)

    def similarity(self: "WordEmbedding", word1: str, word2: str) -> float:
        """میزان شباهت دو کلمه را برمی‌گرداند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('word2vec.bin')
            >>> wordEmbedding.similarity('ایران', 'آلمان')
            0.72231203
            >>> wordEmbedding.similarity('ایران', 'پنجره')
            0.04535884

        Args:
            word1: کلمهٔ اول
            word2: کلمهٔ دوم

        Returns:
            میزان شباهت دو کلمه.

        """
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)

        return float(str(self.model.similarity(word1, word2)))

    def nearest_words(
        self: "WordEmbedding",
        word: str,
        topn: int = 5,
    ) -> List[Tuple[str, str]]:
        """کلمات مرتبط با یک واژه را به همراه میزان ارتباط آن برمی‌گرداند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('word2vec.bin')
            >>> wordEmbedding.nearest_words('ایران', topn = 5)
            [('ایران،', 0.8742443919181824), ('کشور', 0.8735059499740601), ('کشورمان', 0.8443885445594788), ('ایران\u200cبه', 0.8271722197532654), ('خاورمیانه', 0.8266966342926025)]


        Args:
                word: کلمه‌ای که می‌خواهید واژگان مرتبط با آن را بدانید.
                topn: تعداد کلمات مرتبطی که می‌خواهید برگردانده شود.

        Returns:
            لیستی از تاپل‌های [`کلمهٔ مرتبط`, `میزان ارتباط`].

        """
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)
        return self.model.most_similar(word, topn=topn)

    def get_normal_vector(self: "WordEmbedding", word: str) -> Type[ndarray]:
        """بردار امبدینگ نرمالایزشدهٔ کلمه ورودی را برمی‌گرداند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('word2vec.bin')
            >>> result = wordEmbedding.get_normal_vector('سرباز')
            >>> isinstance(result, ndarray)
            True

        Args:
            word: کلمه‌ای که می‌خواهید بردار متناظر با آن را بدانید.

        Returns:
            لیست بردار نرمالایزشدهٔ مرتبط با کلمهٔ ورودی.

        """
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)

        return self.model.get_vector(word=word, norm=True)

    def get_vocabs(self: "WordEmbedding") -> List[str]:
        """لیستی از کلمات موجود در فایل امبدینگ را برمی‌گرداند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('word2vec.bin')
            >>> wordEmbedding.get_vocabs() # doctest: +ELLIPSIS
            ['و', '.', 'در', '،', ...]

        Returns:
            لیست کلمات موجود در فایل امبدینگ.

        """
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)
        return self.model.index_to_key

    def get_vocab_to_index(self: "WordEmbedding") -> dict:
        """دیکشنری برمی‌گرداند که هر کلمه موجود در فایل امبدینگ را به ایندکس آن کلمه در لیست بردارها مپ می‌کند.


        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('word2vec.bin)
            >>> vocab_to_index = wordEmbedding.get_vocab_to_index()
            >>> index = vocab_to_index['سلام']
            >>> vocabs = wordEmbedding.get_vocabs()
            >>> vocabs[index]
            'سلام'

        Returns:
            دیکشنری که هر کلمه را به ایندکس آن مپ می‌کند.
        """
        return self.model.key_to_index

    def get_vectors(self: "WordEmbedding") -> Type[ndarray]:
        """وکتورهای توصیف کننده کلمات را برمیگرداند.(عناصر این وکتور با وکتور کلمات تابع  get_vocabs هم‌اندیس هستند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('resorces/word2vec.bin')
            >>> vectors = wordEmbedding.get_vectors()
            >>> all(vectors[wordEmbedding.get_vocab_to_index()['سلام']] == wordEmbedding['سلام'])
            True

        Returns:
            تمامی وکتور بیان‌کننده کلمات.
        """
        return self.model.vectors

    def get_vector_size(self: "WordEmbedding") -> int:
        """طول وکتور بیان‌کننده هر کلمه در مدل را برمی‌گرداند.

        Examples:
            >>> wordEmbedding = WordEmbedding(model_type = 'fasttext')
            >>> wordEmbedding.load_model('resorces/word2vec.bin')
            >>> wordEmbedding.get_vector_size()
            300


        Returns:
            طول وکتور بیان‌کننده کلمات.

        """
        return self.model.vector_size


class SentenceEmbeddingCorpus:
    """SentenceEmbeddingCorpus."""

    def __init__(self: "SentenceEmbeddingCorpus", data_path: str) -> None:
        """__init__."""
        self.data_path = data_path

    def __iter__(self: "SentenceEmbeddingCorpus") -> Iterator[TaggedDocument]:
        """__iter__."""
        for i, list_of_words in enumerate(smart_open.open(self.data_path)):
            yield TaggedDocument(
                word_tokenize(Normalizer().normalize(list_of_words)),
                [i],
            )

class CallbackSentEmbedding(CallbackAny2Vec):
    def __init__(self: "CallbackSentEmbedding") -> None:
        self.epoch = 0

    def on_epoch_end(self: "CallbackSentEmbedding", model: Doc2Vec):
        print(f"Epoch {self.epoch+1} of {model.epochs}...")
        self.epoch += 1


class SentEmbedding:
    """این کلاس شامل توابعی برای تبدیل جمله به برداری از اعداد است.

    Args:
        model_path: مسیر فایل امبدینگ.

    """

    def __init__(self: "SentEmbedding", model_path: Optional[str] = None) -> None:
        if model_path:
            self.load_model(model_path)
            self.__load_word_embedding_model()

    def __load_word_embedding_model(self: "SentEmbedding") -> None:
        self.word_embedding = WordEmbedding(model_type="keyedvector")
        self.word_embedding.model = self.model.wv

    def load_model(self: "SentEmbedding", model_path: str) -> None:
        """فایل امبدینگ را بارگذاری می‌کند.

        Examples:
            >>> sentEmbedding = SentEmbedding()
            >>> sentEmbedding.load_model('sent2vec.model') # doctest: +ELLIPSIS
            ...

        Args:
            model_path: مسیر فایل امبدینگ.

        """
        self.model = Doc2Vec.load(model_path)
        self.__load_word_embedding_model()

    def train(
        self: "SentEmbedding",
        dataset_path: str,
        min_count: int = 5,
        workers: int = multiprocessing.cpu_count() - 1,  # noqa: B008
        windows: int = 5,
        vector_size: int = 300,
        epochs: int = 10,
        dest_path: str = "gensim_sent2vec.model",
    ) -> None:
        """یک فایل امبدینگ doc2vec ترین می‌کند.

        Examples:
            >>> sentEmbedding = SentEmbedding()
            >>> sentEmbedding.train(dataset_path = 'dataset.txt', min_count = 10, workers = 6, windows = 3, vector_size = 250, epochs = 35, dest_path = 'doc2vec_model') # doctest: +ELLIPSIS
            ...

        Args:
            dataset_path: مسیر فایل متنی.
            min_count: حداقل تعداد تکرار یک کلمه برای قرارگیری آن در مدل امبدینگ.
            workers: تعداد هسته درگیر برای ترین مدل.
            windows: طول پنجره برای لحاظ کلمات اطراف یک کلمه در ترین آن.
            vector_size: طول وکتور خروجی به ازای هر جمله.
            epochs: تعداد تکرار ترین بر روی کل دیتا.
            dest_path: مسیر مورد نظر برای ذخیره فایل امبدینگ.

        """
        workers = 1 if workers == 0 else workers

        doc = SentenceEmbeddingCorpus(dataset_path)

        model = Doc2Vec(
            min_count=min_count,
            window=windows,
            vector_size=vector_size,
            workers=workers,
        )
        print("building vocab...")
        model.build_vocab(doc)
        print("training model...")
        callbacks = [CallbackSentEmbedding()]
        model.train(doc, total_examples=model.corpus_count, epochs=epochs, callbacks=callbacks)

        model.dv.vectors = np.array([[]])
        self.model = model
        self.__load_word_embedding_model()
        print("Model trained.")

        print("saving model...")
        model.save(dest_path)
        print("Model saved.")

    def __getitem__(self: "SentEmbedding", sent: str) -> Type[ndarray]:
        """__getitem__."""
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)
        return self.get_sentence_vector(sent)

    def get_sentence_vector(self: "SentEmbedding", sent: str) -> ndarray:
        """جمله‌ای را دریافت می‌کند و بردار امبدینگ متناظر با آن را برمی‌گرداند.

        Examples:
            >>> sentEmbedding = SentEmbedding()
            >>> sentEmbedding.load_model("sent2vec.model")
            >>> result = sentEmbedding.get_sentence_vector('این متن به برداری متناظر با خودش تبدیل خواهد شد')
            >>> isinstance(result, ndarray)
            True

        Args:
            sent: جمله‌ای که می‌خواهید بردار امبیدنگ آن را دریافت کنید.

        Returns:
            لیست بردار مرتبط با جملهٔ ورودی.

        """
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)

        tokenized_sent = word_tokenize(sent)
        return self.model.infer_vector(tokenized_sent)

    def similarity(self: "SentEmbedding", sent1: str, sent2: str) -> float:
        """میزان شباهت دو جمله را برمی‌گرداند.

        Examples:
            >>> sentEmbedding = SentEmbedding()
            >>> sentEmbedding.load_model("sent2vec.model")
            >>> result = sentEmbedding.similarity('شیر حیوانی وحشی است', 'پلنگ از دیگر جانوران درنده است')
            >>> isinstance(result, float)
            True
            >>> result = sentEmbedding.similarity('هضم یک محصول پردازش متن فارسی است', 'شیر حیوانی وحشی است')
            >>> isinstance(result, float)
            True

        Args:
            sent1: جملهٔ اول.
            sent2: جملهٔ دوم.

        Returns:
            میزان شباهت دو جمله که عددی بین `0` و`1` است.

        """
        if not self.model:
            msg = "Model must not be None! Please load model first."
            raise AttributeError(msg)

        return float(
            str(
                self.model.similarity_unseen_docs(
                    word_tokenize(sent1),
                    word_tokenize(sent2),
                ),
            ),
        )

    def get_vector_size(self: "WordEmbedding") -> int:
        """طول وکتور بیان‌کننده هر جمله در مدل را برمی‌گرداند.

        Examples:
            >>> sentEmbedding = SentEmbedding()
            >>> sentEmbedding.load_model("sent2vec.model")
            >>> sentEmbedding.get_vector_size()
            300


        Returns:
            طول وکتور بیان‌کننده جملات.

        """
        return self.model.vector_size



# ruff: noqa

"""init."""
from .bijankhan_reader import BijankhanReader
from .dadegan_reader import DadeganReader
from .universal_dadegan_reader import UniversalDadeganReader
from .degarbayan_reader import DegarbayanReader
from .hamshahri_reader import HamshahriReader
from .mirastext_reader import MirasTextReader
from .persica_reader import PersicaReader
from .peykare_reader import PeykareReader
from .quran_reader import QuranReader
from .sentipers_reader import SentiPersReader
from .tnews_reader import TNewsReader
from .treebank_reader import TreebankReader
from .verbvalency_reader import VerbValencyReader
from .wikipedia_reader import WikipediaReader



"""این ماژول شامل کلاس‌ها و توابعی برای برچسب‌گذاری توکن‌هاست."""

import time
from typing import List
from typing import Tuple

import numpy as np
from pycrfsuite import Tagger
from pycrfsuite import Trainer
from sklearn.metrics import accuracy_score


def features(sent, index):
    """فهرست فیچرها را برمی‌گرداند."""
    return {
        "word": sent[index],
        "is_first": index == 0,
        "is_last": index == len(sent),
        "is_num": sent[index].isdigit(),
        "prev_word": sent[index - 1] if index != 0 else "",
        "next_word": sent[index + 1] if index != len(sent) - 1 else "",
    }


def data_maker(tokens):
    """تابع دیتا میکر."""
    return [[features(sent, index) for index in range(len(sent))] for sent in tokens]


def iob_features(words, pos_tags, index):
    """تابع iob_features."""
    word_features = features(words, index)
    word_features.update(
        {
            "pos": pos_tags[index],
            "prev_pos": "" if index == 0 else pos_tags[index - 1],
            "next_pos": "" if index == len(pos_tags) - 1 else pos_tags[index + 1],
        },
    )
    return word_features


def iob_data_maker(tokens):
    """تابع iob_data_maker."""
    words = [[word for word, _ in token] for token in tokens]
    tags = [[tag for _, tag in token] for token in tokens]
    return [
        [
            iob_features(words=word_tokens, pos_taggs=tag_tokens, index=index)
            for index in range(len(word_tokens))
        ]
        for word_tokens, tag_tokens in zip(words, tags)
    ]


class SequenceTagger:
    """این کلاس شامل توابعی برای برچسب‌گذاری توکن‌ها است. این کلاس در نقش یک
    wrapper برای کتابخانهٔ [python-crfsuite](https://python-crfsuite.readthedocs.io/en/latest/) است.

    Args:
        model (str, optional): مسیر فایل tagger.
        data_maker (function, optional): تابعی که لیستی دو بعدی از کلمات توکنایز شده را گرفته و لیست دو بعدی از از دیکشنری‌هایی که تعیین‌کننده ویژگی‌ها هر کلمه هستند را برمی‌گرداند.
    """

    def __init__(self: "SequenceTagger", model=None, data_maker=data_maker) -> None:
        if model is not None:
            self.load_model(model)
        else:
            self.model = None
        self.data_maker = data_maker

    def __add_label(self: "SequenceTagger", sentence, tags):
        return [(word, tag) for word, tag in zip(sentence, tags)]

    def __tag(self: "SequenceTagger", tokens):
        return self.__add_label(tokens, self.model.tag(self.data_maker([tokens])[0]))

    def __train(
        self: "SequenceTagger", x, y, args, verbose, file_name, report_duration,
    ):
        trainer = Trainer(verbose=verbose)
        trainer.set_params(args)

        for xseq, yseq in zip(x, y):
            trainer.append(xseq, yseq)

        start_time = time.time()
        trainer.train(file_name)
        end_time = time.time()

        if report_duration:
            print(f"training time: {(end_time - start_time):.2f} sec")

        self.load_model(file_name)

    def __evaluate(self: "SequenceTagger", gold_labels, predicted_labels):
        gold_labels = np.concatenate(gold_labels)
        predicted_labels = np.concatenate(predicted_labels)
        return float(accuracy_score(gold_labels, predicted_labels))

    def load_model(self: "SequenceTagger", model):
        """فایل تگر را بارگزاری می‌کند.

        Examples:
            >>> tagger = SequenceTagger()
            >>> tagger.load_model(model = 'tagger.model')

        Args:
            model (str): مسیر فایل تگر.

        """
        tagger = Tagger()
        tagger.open(model)
        self.model = tagger

    def tag(self: "SequenceTagger", tokens):
        """یک جمله را در قالب لیستی از توکن‌ها دریافت می‌کند و در خروجی لیستی از
        `(توکن، برچسب)`ها برمی‌گرداند.

        Examples:
            >>> tagger = SequenceTagger(model = 'tagger.model')
            >>> tagger.tag(tokens = ['من', 'به', 'مدرسه', 'ایران', 'رفته_بودم', '.'])
            [('من', 'PRON'), ('به', 'ADP'), ('مدرسه', 'NOUN,EZ'), ('ایران', 'NOUN'), ('رفته_بودم', 'VERB'), ('.', 'PUNCT')]

        Args:
            tokens (List[str]): لیستی از توکن‌های یک جمله که باید برچسب‌گذاری شود.

        Returns:
            (List[Tuple[str,str]]): ‌لیستی از `(توکن، برچسب)`ها.

        """
        if self.model is None:
            msg = "you should load model first..."
            raise ValueError(msg)

        return self.__tag(tokens)

    def tag_sents(self: "SequenceTagger", sentences):
        """جملات را در قالب لیستی از توکن‌ها دریافت می‌کند
        و در خروجی، لیستی از لیستی از `(توکن، برچسب)`ها برمی‌گرداند.

        هر لیست از `(توکن، برچسب)`ها مربوط به یک جمله است.

        Examples:
            >>> tagger = SequenceTagger(model = 'tagger.model')
            >>> tagger.tag_sents(sentences = [['من', 'به', 'مدرسه', 'ایران', 'رفته_بودم', '.']])
            [[('من', 'PRON'), ('به', 'ADP'), ('مدرسه', 'NOUN,EZ'), ('ایران', 'NOUN'), ('رفته_بودم', 'VERB'), ('.', 'PUNCT')]]

        Args:
            sentences (List[List[str]]): لیستی از جملات که باید برچسب‌گذاری شود.

        Returns:
            (List[List[Tuple[str,str]]]): لیستی از لیستی از `(توکن، برچسب)`ها.
                    هر لیست از `(توکن،برچسب)`ها مربوط به یک جمله است.

        """
        if self.model is None:
            msg = "you should load model first..."
            raise ValueError(msg)

        return [self.__tag(tokens) for tokens in sentences]

    def train(
        self: "SequenceTagger",
        tagged_list,
        c1=0.4,
        c2=0.04,
        max_iteration=400,
        verbose=True,
        file_name="crf.model",
        report_duration=True,
    ):
        """لیستی از جملات را می‌گیرد و بر اساس آن مدل را آموزش می‌دهد.

        هر جمله، لیستی از `(توکن، برچسب)`هاست.

        Examples:
            >>> tagger = SequenceTagger()
            >>> tagger.train(tagged_list = [[('من', 'PRON'), ('به', 'ADP'), ('مدرسه', 'NOUN,EZ'), ('ایران', 'NOUN'), ('رفته_بودم', 'VERB'), ('.', 'PUNCT')]], c1 = 0.5, c2 = 0.5, max_iteration = 100, verbose = True, file_name = 'tagger.model', report_duration = True)
            Feature generation
            type: CRF1d
            feature.minfreq: 0.000000
            feature.possible_states: 0
            feature.possible_transitions: 1
            0....1....2....3....4....5....6....7....8....9....10
            Number of features: 150
            Seconds required: 0.001
            ...
            Writing feature references for attributes
            Seconds required: 0.000

            training time: 0.01 sec

        Args:
            tagged_list (List[{List[Tuple[str,str]]]): جملاتی که مدل از روی آن‌ها آموزش می‌بیند.
            c1 (float): مقدار L1 regularization.
            c2 (float): مقدار L2 regularization.
            max_iteration (int): تعداد تکرار آموزش بر کل دیتا.
            verbose (boolean): نمایش اطلاعات مربوط به آموزش.
            file_name (str): نام و مسیر فایلی که می‌خواهید مدل در آن ذخیره شود.
            report_duration (boolean): نمایش گزارشات مربوط به زمان.

        """
        x = self.data_maker(
            [[x for x, _ in tagged_sent] for tagged_sent in tagged_list],
        )
        y = [[y for _, y in tagged_sent] for tagged_sent in tagged_list]

        args = {
            "c1": c1,
            "c2": c2,
            "max_iterations": max_iteration,
            "feature.possible_transitions": True,
        }

        self.__train(x, y, args, verbose, file_name, report_duration)

    def save_model(self: "SequenceTagger", filename):
        """مدل تهیه‌شده توسط تابع [train()][hazm.sequence_tagger.SequenceTagger.train]
        را ذخیره می‌کند.

        Examples:
            >>> tagger = SequenceTagger()
            >>> tagger.train(tagged_list = [[('من', 'PRON'), ('به', 'ADP'), ('مدرسه', 'NOUN,EZ'), ('ایران', 'NOUN'), ('رفته_بودم', 'VERB'), ('.', 'PUNCT')]], c1 = 0.5, c2 = 0.5, max_iteration = 100, verbose = True, file_name = 'tagger.model', report_duration = True)
            >>> tagger.save_model(file_name = 'new_tagger.model')

        Args:
            filename (str): نام و مسیر فایلی که می‌خواهید مدل در آن ذخیره شود.

        """
        if self.model is None:
            msg = "you should load model first..."
            raise ValueError(msg)

        self.model.dump(filename)

    def evaluate(
        self: "SequenceTagger", tagged_sent: List[List[Tuple[str, str]]],
    ) -> float:
        """داده صحیح دریافت شده را با استفاده از مدل لیبل می‌زند و دقت مدل را برمی‌گرداند.

        Examples:
            >>> tagger = SequenceTagger(model = 'tagger.model')
            >>> tagger.evaluate([[('نامه', 'NOUN,EZ'), ('ایشان', 'PRON'), ('را', 'ADP'), ('دریافت', 'NOUN'), ('داشتم', 'VERB'), ('.', 'PUNCT')]])
            1.0


        Args:
            tagged_sent: جملات لیبل‌داری که با استفاده از آن مدل را ارزیابی می‌کنیم.

        Returns:
            دقت مدل

        """
        if self.model is None:
            msg = "you should load model first..."
            raise ValueError(msg)

        def extract_labels(tagged_list):
            return [[label for _, label in sent] for sent in tagged_list]

        tokens = [[word for word, _ in sent] for sent in tagged_sent]
        predicted_tagged_sent = self.tag_sents(tokens)
        return self.__evaluate(
            extract_labels(tagged_sent),
            extract_labels(predicted_tagged_sent),
        )


class IOBTagger(SequenceTagger):
    """IOBTagger."""

    def __init__(self: "SequenceTagger", model=None, data_maker=iob_data_maker) -> None:
        super().__init__(model, data_maker)

    def __iob_format(self: "SequenceTagger", tagged_data, chunk_tags):
        return [
            (token[0], token[1], chunk_tag[1])
            for token, chunk_tag in zip(tagged_data, chunk_tags)
        ]

    def tag(self: "SequenceTagger", tagged_data):
        """یک جمله را در قالب لیستی از توکن‌ها و تگ‌ها دریافت می‌کند و در خروجی لیستی از
        `(توکن، تگ، برچسب)`ها برمی‌گرداند.

        Examples:
            >>> iobTagger = IOBTagger(model = 'tagger.model')
            >>> iobTagger.tag(tagged_data = [('من', 'PRON'), ('به', 'ADP'), ('مدرسه', 'NOUN,EZ'), ('ایران', 'NOUN'), ('رفته_بودم', 'VERB'), ('.', 'PUNCT')])
            [('من', 'PRON', 'B-NP'), ('به', 'ADP', 'B-PP'), ('مدرسه', 'NOUN,EZ', 'B-NP'), ('ایران', 'NOUN', 'I-NP'), ('رفته_بودم', 'VERB', 'B-VP'), ('.', 'PUNCT', 'O')]

        Args:
            tagged_data (List[Tuple[str, str]]): لیستی از توکن‌های یک جمله که باید برچسب‌گذاری شود.

        Returns:
            (List[Tuple[str, str, str]]): ‌لیستی از `(توکن، تگ، برچسب)`ها.

        """
        chunk_tags = super().tag(tagged_data)
        return self.__iob_format(tagged_data, chunk_tags)

    def tag_sents(self: "SequenceTagger", sentences):
        """جملات را در قالب لیستی لیستی از توکن‌ها و تگ‌ها دریافت می‌کند
        و در خروجی، لیستی از لیستی از `(توکن، تگ، برچسب)`ها برمی‌گرداند.

        هر لیست از `(توکن، برچسب)`ها مربوط به یک جمله است.

        Examples:
            >>> iobTagger = IOBTagger(model = 'tagger.model')
            >>> iobTagger.tag_sents(tagged_data = [[('من', 'PRON'), ('به', 'ADP'), ('مدرسه', 'NOUN,EZ'), ('ایران', 'NOUN'), ('رفته_بودم', 'VERB'), ('.', 'PUNCT')]])
            [[('من', 'PRON', 'B-NP'), ('به', 'ADP', 'B-PP'), ('مدرسه', 'NOUN,EZ', 'B-NP'), ('ایران', 'NOUN', 'I-NP'), ('رفته_بودم', 'VERB', 'B-VP'), ('.', 'PUNCT', 'O')]]

        Args:
            sentences (List[List[str]]): لیستی از جملات که باید برچسب‌گذاری شود.

        Returns:
            (List[List[Tuple[str,str]]]): لیستی از لیستی از `(توکن، تگ، برچسب)`ها.
                    هر لیست از `(توکن، تگ، برچسب)`ها مربوط به یک جمله است.

        """
        chunk_tags = super().tag_sents(sentences)
        return [
            self.__iob_format(tagged_data, chunks)
            for tagged_data, chunks in zip(sentences, chunk_tags)
        ]

    def train(
        self: "SequenceTagger",
        tagged_list: List[List[Tuple[str, str, str]]],
        c1: float = 0.4,
        c2: float = 0.04,
        max_iteration: int = 400,
        verbose: True = True,
        file_name: str = "crf.model",
        report_duration=True,
    ):
        """لیستی از جملات را می‌گیرد و بر اساس آن مدل را آموزش می‌دهد.

        هر جمله، لیستی از `(توکن، تگ، برچسب)`هاست.

        Examples:
            >>> iobTagger = IOBTagger()
            >>> iobTagger.train(tagged_list = [[('من', 'PRON', 'B-NP'), ('به', 'ADP', 'B-PP'), ('مدرسه', 'NOUN,EZ', 'B-NP'), ('ایران', 'NOUN', 'I-NP'), ('رفته_بودم', 'VERB', 'B-VP'), ('.', 'PUNCT', 'O')]], c1 = 0.5, c2 = 0.5, max_iteration = 100, verbose = True, file_name = 'newIOBTagger.model', report_duration = True)
            Feature generation
            type: CRF1d
            feature.minfreq: 0.000000
            feature.possible_states: 0
            feature.possible_transitions: 1
            0....1....2....3....4....5....6....7....8....9....10
            Number of features: 150
            Seconds required: 0.001
            ...
            Writing feature references for attributes
            Seconds required: 0.000

            training time: 0.01 sec

        Args:
            tagged_list: جملاتی که مدل از روی آن‌ها آموزش می‌بیند.
            c1: مقدار L1 regularization.
            c2: مقدار L2 regularization.
            max_iteration: تعداد تکرار آموزش بر کل دیتا.
            verbose: نمایش اطلاعات مربوط به آموزش.
            file_name: نام و مسیر فایلی که می‌خواهید مدل در آن ذخیره شود.
            report_duration: نمایش گزارشات مربوط به زمان.

        """
        tagged_list = [
            [((word, tag), chunk) for word, tag, chunk in tagged_sent]
            for tagged_sent in tagged_list
        ]
        return super().train(
            tagged_list,
            c1,
            c2,
            max_iteration,
            verbose,
            file_name,
            report_duration,
        )

    def evaluate(
        self: "SequenceTagger", tagged_sent: List[List[Tuple[str, str, str]]],
    ) -> float:
        """داده صحیح دریافت شده را با استفاده از مدل لیبل می‌زند و دقت مدل را برمی‌گرداند.

        Examples:
            >>> iobTagger = IOBTagger(model = 'tagger.model')
            >>> iobTagger.evaluate([[('نامه', 'NOUN,EZ', 'B-NP'), ('ایشان', 'PRON', 'I-NP'), ('را', 'ADP', 'B-POSTP'), ('دریافت', 'NOUN', 'B-VP'), ('داشتم', 'VERB', 'I-VP'), ('.', 'PUNCT', 'O')]])
            1.0


        Args:
            tagged_sent: جملات لیبل‌داری که با استفاده از آن مدل را ارزیابی می‌کنیم.

        Returns:
            دقت مدل

        """
        if self.model is None:
            msg = "you should load model first..."
            raise ValueError(msg)

        def extract_labels(tagged_list):
            return [[token[-1] for token in sent] for sent in tagged_list]

        tokens = [[token[:-1] for token in sent] for sent in tagged_sent]
        predicted_tagged_sent = self.tag_sents(tokens)
        return super()._SequenceTagger__evaluate(
            extract_labels(tagged_sent),
            extract_labels(predicted_tagged_sent),
        )



"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ سِنتی‌پِرِس است.

سِنتی‌پرس شامل مجموعه‌ای از متون فارسی با برچسب‌های معنایی است.

"""


import itertools
import os
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from xml.dom import minidom


class SentiPersReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ سِنتی‌پِرِس است.

    Args:
        root: مسیر فولدر حاوی فایل‌های پیکره

    """

    def __init__(self: "SentiPersReader", root: str) -> None:
        self._root = root

    def docs(self: "SentiPersReader") -> Iterator[Dict[str, Any]]:
        """متن‌های فارسی را در قالب یک برمی‌گرداند.

        هر متن شامل این فیلدهاست:

        - عنوان (Title)
        - نوع (Type)
        - نظرات (comments)

        فیلد `comments `خودش شامل این فیلدهاست:

        - شناسه (id)
        - نوع (type)
        - نویسنده (author)
        - ارزش (value)
        - جملات (sentences)

        Yields:
            متن بعدی.

        """

        def element_sentences(element: str) -> Iterator[Dict]:
            for sentence in element.getElementsByTagName("Sentence"):
                yield {
                    "text": sentence.childNodes[0].data,
                    "id": sentence.getAttribute("ID"),
                    "value": (
                        int(sentence.getAttribute("Value"))
                        if comment.getAttribute("Value")
                        else None
                    ),
                }

        for root, _dirs, files in os.walk(self._root):
            for filename in sorted(files):
                try:
                    elements = minidom.parse(os.path.join(root, filename)) # noqa: PTH118

                    product = elements.getElementsByTagName("Product")[0]
                    doc = {
                        "Title": product.getAttribute("Title"),
                        "Type": product.getAttribute("Type"),
                        "comments": [],
                    }

                    for child in product.childNodes:
                        if child.nodeName in {
                            "Voters",
                            "Performance",
                            "Capability",
                            "Production_Quality",
                            "Ergonomics",
                            "Purchase_Value",
                        }:
                            value = child.getAttribute("Value")
                            doc[child.nodeName] = (
                                float(value) if "." in value else int(value)
                            )

                    for comment in itertools.chain(
                        elements.getElementsByTagName("Opinion"),
                        elements.getElementsByTagName("Criticism"),
                    ):
                        doc["comments"].append(
                            {
                                "id": comment.getAttribute("ID"),
                                "type": comment.nodeName,
                                "author": comment.getAttribute("Holder").strip(),
                                "value": (
                                    int(comment.getAttribute("Value"))
                                    if comment.getAttribute("Value")
                                    else None
                                ),
                                "sentences": list(element_sentences(comment)),
                            },
                        )

                    # todo: Accessories, Features, Review, Advantages, Tags, Keywords, Index

                    yield doc

                except Exception as e:
                    print("error in reading", filename, e, file=sys.stderr)

    def comments(self: "SentiPersReader") -> Iterator[str]:
        """نظرات مربوط به متن را برمی‌گرداند.

        Examples:
            >>> sentipers = SentiPersReader(root='sentipers')
            >>> next(sentipers.comments())[0][1]
            'بيشتر مناسب است براي کساني که به دنبال تنوع هستند و در همه چيز نو گرايي دارند .'

        Yields:
            نظر بعدی.

        """
        for doc in self.docs():
            yield [
                [sentence["text"] for sentence in text]
                for text in [comment["sentences"] for comment in doc["comments"]]
            ]



"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ PerDT است.

PerDT حاوی تعداد قابل‌توجهی جملۀ برچسب‌خورده با اطلاعات نحوی و ساخت‌واژی است.

"""
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Type

from nltk.parse import DependencyGraph
from nltk.tree import Tree


def coarse_pos_u(tags: List[str], word: str) -> str:
    """برچسب‌های ریز را به برچسب‌های درشت منطبق با استاندارد جهانی (coarse-grained
    universal pos tags) تبدیل می‌کند.

    Examples:
        >>> coarse_pos_e(['N', 'IANM'], 'امروز')
        'N'

    """
    mapping = {
        "N": "NOUN",
        "V": "VERB",
        "ADJ": "ADJ",
        "ADV": "ADV",
        "PR": "PRON",
        "PREM": "DET",
        "PREP": "ADP",
        "POSTP": "ADP",
        "PRENUM": "NUM",
        "CONJ": "CCONJ",
        "PUNC": "PUNCT",
        "SUBR": "SCONJ",
        "IDEN": "PROPN",
        "POSTNUM": "NUM",
        "PSUS": "INTJ",
        "PART": "PART",
        "ADR": "INTJ",
    }
    pos_mapped = mapping.get(tags[0], "X")
    if pos_mapped == "PART" and word == "را":
        return "ADP"
    if pos_mapped == "PART" and word in ["خوب", "آخر"]:
        return "ADP"
    return pos_mapped


def coarse_pos_e(tags: List[str], word) -> str: # noqa: ARG001
    """برچسب‌های ریز را به برچسب‌های درشت (coarse-grained pos tags) تبدیل می‌کند.

    Examples:
        >>> coarse_pos_e(['N', 'IANM'],'امروز')
        'N'

    """
    mapping = {
        "N": "N",
        "V": "V",
        "ADJ": "AJ",
        "ADV": "ADV",
        "PR": "PRO",
        "PREM": "DET",
        "PREP": "P",
        "POSTP": "POSTP",
        "PRENUM": "NUM",
        "CONJ": "CONJ",
        "PUNC": "PUNC",
        "SUBR": "CONJ",
    }
    return mapping.get(tags[0], "X") + ("e" if "EZ" in tags else "")


def word_nodes(tree: Type[Tree]) -> List[Dict[str, Any]]:
    """نودها را به صورت مرتب‌شده برمی‌گرداند."""
    return sorted(tree.nodes.values(), key=lambda node: node["address"])[1:]


def node_deps(node: List[Dict[str, Any]]) -> List[Any]:
    """مقادیر موجود در فیلد deps نود ورودی را برمی‌گرداند."""
    return sum(list(node["deps"].values()), [])


class DadeganReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ PerDT است.

    Args:
        conll_file: مسیر فایلِ پیکره.
        pos_map: دیکشنری مبدل برچسب‌های ریز به درشت.

    """

    def __init__(
        self: "DadeganReader",
        conll_file: str,
        pos_map: str = coarse_pos_e,
        universal_pos: bool = False,
    ) -> None:
        self._conll_file = conll_file
        if pos_map is None:
            self._pos_map = lambda tags: ",".join(tags)
        elif universal_pos:
            self._pos_map = coarse_pos_u
        else:
            self._pos_map = coarse_pos_e

    def _sentences(self: "DadeganReader") -> Iterator[str]:
        """جملات پیکره را به شکل متن خام برمی‌گرداند.

        Yields:
            جملهٔ بعدی.

        """
        with Path(self._conll_file).open(encoding="utf8") as conll_file:
            text = conll_file.read()

            # refine text
            text = (
                text.replace("‌‌", "‌")
                .replace("\t‌", "\t")
                .replace("‌\t", "\t")
                .replace("\t ", "\t")
                .replace(" \t", "\t")
                .replace("\r", "")
                .replace("\u2029", "‌")
            )

            for item in text.replace(" ", "_").split("\n\n"):
                if item.strip():
                    yield item

    def trees(self: "DadeganReader") -> Iterator[Type[Tree]]:
        """ساختار درختی جملات را برمی‌گرداند.

        Yields:
            ساختار درختی جملهٔ بعدی.

        """
        for sentence in self._sentences():
            tree = DependencyGraph(sentence)

            for node in word_nodes(tree):
                node["mtag"] = [node["ctag"], node["tag"]]

                if "ezafe" in node["feats"]:
                    node["mtag"].append("EZ")

                node["mtag"] = self._pos_map(node["mtag"], node["word"])

            yield tree

    def sents(self: "DadeganReader") -> Iterator[List[Tuple[str, str]]]:
        """لیستی از جملات را برمی‌گرداند.

        هر جمله لیستی از `(توکن، برچسب)`ها است.

        Examples:
            >>> dadegan = DadeganReader(conll_file='dadegan.conll')
            >>> next(dadegan.sents())
            [('این', 'DET'), ('میهمانی', 'N'), ('به', 'P'), ('منظور', 'Ne'), ('آشنایی', 'Ne'), ('هم‌تیمی‌های', 'Ne'), ('او', 'PRO'), ('با', 'P'), ('غذاهای', 'Ne'), ('ایرانی', 'AJ'), ('ترتیب', 'N'), ('داده_شد', 'V'), ('.', 'PUNC')]

        Yields:
            جملهٔ بعدی.

        """
        for tree in self.trees():
            yield [(node["word"], node["mtag"]) for node in word_nodes(tree)]

    def chunked_trees(self: "DadeganReader") -> Iterator[Type[Tree]]:
        """درخت وابستگی‌های جملات را برمی‌گرداند.

        Examples:
            >>> from hazm.chunker import tree2brackets
            >>> dadegan = DadeganReader(conll_file='dadegan.conll')
            >>> tree2brackets(next(dadegan.chunked_trees()))
            '[این میهمانی NP] [به PP] [منظور آشنایی هم‌تیمی‌های او NP] [با PP] [غذاهای ایرانی NP] [ترتیب داده_شد VP] .'

        Yields:
            درخت وابستگی‌های جملهٔ بعدی.

        """
        for tree in self.trees():
            chunks = []
            for node in word_nodes(tree):
                n = node["address"]
                item = (node["word"], node["mtag"])
                appended = False
                if node["ctag"] in {"PREP", "POSTP"}:
                    for d in node_deps(node):
                        label = "PP"
                        if node["ctag"] == "POSTP":
                            label = "POSTP"
                        if (
                            d == n - 1
                            and type(chunks[-1]) == Tree
                            and chunks[-1].label() == label
                        ):
                            chunks[-1].append(item)
                            appended = True
                    if (
                        node["head"] == n - 1
                        and len(chunks) > 0
                        and type(chunks[-1]) == Tree
                        and chunks[-1].label() == label
                    ):
                        chunks[-1].append(item)
                        appended = True
                    if not appended:
                        chunks.append(Tree(label, [item]))
                elif node["ctag"] in {"PUNC", "CONJ", "SUBR", "PART"}:
                    if (
                        item[0]
                        in {"'", '"', "(", ")", "{", "}", "[", "]", "-", "#", "«", "»"}
                        and len(chunks) > 0
                        and type(chunks[-1]) == Tree
                    ):
                        for leaf in chunks[-1].leaves():
                            if leaf[1] == item[1]:
                                chunks[-1].append(item)
                                appended = True
                                break
                    if appended is not True:
                        chunks.append(item)
                elif node["ctag"] in {
                    "N",
                    "PREM",
                    "ADJ",
                    "PR",
                    "ADR",
                    "PRENUM",
                    "IDEN",
                    "POSNUM",
                    "SADV",
                }:
                    if node["rel"] in {"MOZ", "NPOSTMOD"}:
                        if len(chunks) > 0:
                            if type(chunks[-1]) == Tree:
                                j = n - len(chunks[-1].leaves())
                                chunks[-1].append(item)
                            else:
                                j = n - 1
                                treenode = Tree("NP", [chunks.pop(), item])
                                chunks.append(treenode)
                            while j > node["head"]:
                                leaves = chunks.pop().leaves()
                                if len(chunks) < 1:
                                    chunks.append(Tree("NP", leaves))
                                    j -= 1
                                elif type(chunks[-1]) == Tree:
                                    j -= len(chunks[-1])
                                    for leaf in leaves:
                                        chunks[-1].append(leaf)
                                else:
                                    leaves.insert(0, chunks.pop())
                                    chunks.append(Tree("NP", leaves))
                                    j -= 1
                            continue
                    elif node["rel"] == "POSDEP" and tree.nodes[node["head"]][
                        "rel"
                    ] in {"NCONJ", "AJCONJ"}:
                        conj = tree.nodes[node["head"]]
                        if tree.nodes[conj["head"]]["rel"] in {
                            "MOZ",
                            "NPOSTMOD",
                            "AJCONJ",
                            "POSDEP",
                        }:
                            label = "NP"
                            leaves = [item]
                            j = n - 1
                            while j >= conj["head"]:
                                if type(chunks[-1]) is Tree:
                                    j -= len(chunks[-1].leaves())
                                    label = chunks[-1].label()
                                    leaves = chunks.pop().leaves() + leaves
                                else:
                                    leaves.insert(0, chunks.pop())
                                    j -= 1
                            chunks.append(Tree(label, leaves))
                            appended = True
                    elif (
                        node["head"] == n - 1
                        and len(chunks) > 0
                        and type(chunks[-1]) == Tree
                        and chunks[-1].label() != "PP"
                    ):
                        chunks[-1].append(item)
                        appended = True
                    elif node["rel"] == "AJCONJ" and tree.nodes[node["head"]][
                        "rel"
                    ] in {"NPOSTMOD", "AJCONJ"}:
                        np_nodes = [item]
                        label = "ADJP"
                        i = n - node["head"]
                        while i > 0:
                            if type(chunks[-1]) == Tree:
                                label = chunks[-1].label()
                                leaves = chunks.pop().leaves()
                                i -= len(leaves)
                                np_nodes = leaves + np_nodes
                            else:
                                i -= 1
                                np_nodes.insert(0, chunks.pop())
                        chunks.append(Tree(label, np_nodes))
                        appended = True
                    elif (
                        node["ctag"] == "ADJ"
                        and node["rel"] == "POSDEP"
                        and tree.nodes[node["head"]]["ctag"] != "CONJ"
                    ):
                        np_nodes = [item]
                        i = n - node["head"]
                        while i > 0:
                            label = "ADJP"
                            if type(chunks[-1]) == Tree:
                                label = chunks[-1].label()
                                leaves = chunks.pop().leaves()
                                i -= len(leaves)
                                np_nodes = leaves + np_nodes
                            else:
                                i -= 1
                                np_nodes.insert(0, chunks.pop())
                        chunks.append(Tree(label, np_nodes))
                        appended = True
                    for d in node_deps(node):
                        if (
                            d == n - 1
                            and type(chunks[-1]) == Tree
                            and chunks[-1].label() != "PP"
                            and appended is not True
                        ):
                            label = chunks[-1].label()
                            if node["rel"] == "ADV":
                                label = "ADVP"
                            elif label in {"ADJP", "ADVP"}:
                                if node["ctag"] == "N":
                                    label = "NP"
                                elif node["ctag"] == "ADJ":
                                    label = "ADJP"
                            leaves = chunks.pop().leaves()
                            leaves.append(item)
                            chunks.append(Tree(label, leaves))
                            appended = True
                        elif tree.nodes[d]["rel"] == "NPREMOD" and appended is not True:
                            np_nodes = [item]
                            i = n - d
                            while i > 0:
                                if type(chunks[-1]) == Tree:
                                    leaves = chunks.pop().leaves()
                                    i -= len(leaves)
                                    np_nodes = leaves + np_nodes
                                else:
                                    i -= 1
                                    np_nodes.insert(0, chunks.pop())
                            chunks.append(Tree("NP", np_nodes))
                            appended = True
                    if not appended:
                        label = "NP"
                        if node["ctag"] == "ADJ":
                            label = "ADJP"
                        elif node["rel"] == "ADV":
                            label = "ADVP"
                        chunks.append(Tree(label, [item]))
                elif node["ctag"] in {"V"}:
                    appended = False
                    for d in node_deps(node):
                        if (
                            d == n - 1
                            and type(chunks[-1]) == Tree
                            and tree.nodes[d]["rel"] in {"NVE", "ENC"}
                            and appended is not True
                        ):
                            leaves = chunks.pop().leaves()
                            leaves.append(item)
                            chunks.append(Tree("VP", leaves))
                            appended = True
                        elif tree.nodes[d]["rel"] in {"VPRT", "NVE"}:
                            vp_nodes = [item]
                            i = n - d
                            while i > 0:
                                if type(chunks[-1]) == Tree:
                                    leaves = chunks.pop().leaves()
                                    i -= len(leaves)
                                    vp_nodes = leaves + vp_nodes
                                else:
                                    i -= 1
                                    vp_nodes.insert(0, chunks.pop())
                            chunks.append(Tree("VP", vp_nodes))
                            appended = True
                            break
                    if not appended:
                        chunks.append(Tree("VP", [item]))
                elif node["ctag"] in {"PSUS"}:
                    if node["rel"] == "ADV":
                        chunks.append(Tree("ADVP", [item]))
                    else:
                        chunks.append(Tree("VP", [item]))
                elif node["ctag"] in {"ADV", "SADV"}:
                    appended = False
                    for d in node_deps(node):
                        if d == n - 1 and type(chunks[-1]) == Tree:
                            leaves = chunks.pop().leaves()
                            leaves.append(item)
                            chunks.append(Tree("ADVP", leaves))
                            appended = True
                    if not appended:
                        chunks.append(Tree("ADVP", [item]))

            yield Tree("S", chunks)



"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ Quranic Arabic است.

پیکرهٔ [Quranic Arabic](https://corpus.quran.com/) شامل قواعد نحوی و اطلاعات
ریخت‌شناسی تک‌تک کلمات قرآن کریم است.

"""
from pathlib import Path
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple

from hazm.utils import maketrans

buckwalter_transliteration = maketrans(
    "'>&<}AbptvjHxd*rzs$SDTZEg_fqklmnhwYyFNKaui~o^#`{:@\"[;,.!-+%]",
    "\u0621\u0623\u0624\u0625\u0626\u0627\u0628\u0629\u062a\u062b\u062c\u062d\u062e\u062f\u0630\u0631\u0632\u0633\u0634\u0635\u0636\u0637\u0638\u0639\u063a\u0640\u0641\u0642\u0643\u0644\u0645\u0646\u0647\u0648\u0649\u064a\u064b\u064c\u064d\u064e\u064f\u0650\u0651\u0652\u0653\u0654\u0670\u0671\u06dc\u06df\u06e0\u06e2\u06e3\u06e5\u06e6\u06e8\u06ea\u06eb\u06ec\u06ed",
)


class QuranReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ Quranic Arabic است.

    Args:
        quran_file: مسیر فایلِ پیکره

    """

    def __init__(self: "QuranReader", quran_file: str) -> None:
        self._quran_file = quran_file

    def parts(self: "QuranReader") -> Iterator[Dict[str, str]]:
        """اجزای متن قرآن را به‌همراه اطلاعات نحوی‌شان برمی‌گرداند.

        یک جزء لزوماً یک کلمه نیست؛ مثلاً واژهٔ «الرحمن» از دو جزء «ال» و «رحمن» تشکیل
        شده است.

        Examples:
            >>> parts=QuranReader(quran_file='quranic_corpus_morphology.txt').parts()
            >>> print(next(parts))
            {'loc': (1, 1, 1, 1), 'text': 'بِ', 'tag': 'P'}
            >>> print(next(parts))
            {'loc': (1, 1, 1, 2), 'text': 'سْمِ', 'tag': 'N', 'lem': 'ٱسْم', 'root': 'سمو'}
            >>> print(next(parts))
            {'loc': (1, 1, 2, 1), 'text': 'ٱللَّهِ', 'tag': 'PN', 'lem': 'ٱللَّه', 'root': 'اله'}

        Yields:
            جزء بعدی متن قرآن.

        """
        for line in Path(self._quran_file).open(encoding="utf8"):
            if not line.startswith("("):
                continue
            parts = line.strip().split("\t")

            part = {
                "loc": eval(parts[0].replace(":", ",")), # noqa: PGH001
                "text": parts[1].translate(buckwalter_transliteration),
                "tag": parts[2],
            }

            features = parts[3].split("|")
            for feature in features:
                if feature.startswith("LEM:"):
                    part["lem"] = feature[4:].translate(buckwalter_transliteration)
                elif feature.startswith("ROOT:"):
                    part["root"] = feature[5:].translate(buckwalter_transliteration)
            yield part

    def words(
        self: "QuranReader",
    ) -> Iterator[Tuple[str, str, str, str, str, List[Dict[str, str]]]]:
        """اطلاعات صرفی کلمات قرآن را برمی‌گرداند.

        Examples:
            >>> words=QuranReader(quran_file='quranic_corpus_morphology.txt').words()
            >>> print(next(words))
            ('1.1.1', 'بِسْمِ', 'ٱسْم', 'سمو', 'P-N', [{'text': 'بِ', 'tag': 'P'}, {'text': 'سْمِ', 'tag': 'N', 'lem': 'ٱسْم', 'root': 'سمو'}])

        Yields:
            اطلاعات صرفی کلمهٔ بعدی قرآن.

        """

        def word_item(location: Tuple[int], parts: List[Dict]) -> str:
            text = "".join([part["text"] for part in parts])
            tag = "-".join([part["tag"] for part in parts])
            lem = "-".join([part["lem"] for part in parts if "lem" in part])
            root = "-".join([part["root"] for part in parts if "root" in part])
            return ".".join(map(str, location)), text, lem, root, tag, parts

        last_location = (0, 0, 0, 0)
        items = []
        for part in self.parts():
            if last_location[:3] == part["loc"][:3]:
                items.append(part)
            else:
                if items:
                    yield word_item(last_location[:3], items)
                items = [part]
            last_location = part["loc"]
            del part["loc"]
        yield word_item(last_location[:3], items)



"""این ماژول، پیکره‌های متنی خام را می‌خواند."""
from typing import Any
from typing import Callable
from typing import List

from nltk.corpus import PlaintextCorpusReader
from nltk.corpus.reader import StreamBackedCorpusView
from nltk.corpus.reader import read_blankline_block

from hazm import SentenceTokenizer
from hazm import WordTokenizer


class PersianPlainTextReader(PlaintextCorpusReader):
    """Reader for corpora that consist of plaintext documents.  Paragraphs
    are assumed to be split using blank lines.  Sentences and words can
    be tokenized using the default tokenizers, or by custom tokenizers
    specificed as parameters to the constructor.

    """

    CorpusView = StreamBackedCorpusView

    def __init__(
        self: "PersianPlainTextReader",
        root: str,
        fileids: List,
        word_tokenizer: Callable = WordTokenizer.tokenize,
        sent_tokenizer: Callable = SentenceTokenizer.tokenize,
        para_block_reader: Callable = read_blankline_block,
        encoding: str = "utf8",
    ) -> None:
        super().__init__(
            root,
            fileids,
            word_tokenizer,
            sent_tokenizer,
            para_block_reader,
            encoding,
        )



"""این ماژول شامل کلاس‌ها و توابعی برای خواندن پیکرهٔ ظرفیت نحوی افعال فارسی
است.

پیکرهٔ ظرفیت نحوی افعال فارسی مجموعه‌ای است حاوی اطلاعات مربوط به ظرفیت نحوی
بیش
از ۴۵۰۰ فعل در زبان فارسی. در این فرهنگ، متمم‌های اجباری و اختیاری انواع
فعل‌های
ساده، مرکب، پیشوندی و عبارات فعلی مشخص شده است. فراوانی فعل‌های مرکب در زبان
فارسی، نیاز به فرهنگ ظرفیت فعل را در این زبان دوچندان می‌نماید. چرا که شناخت
فعل‌های مرکب چه از لحاظ انسانی و چه از لحاظ پردازشی کاری دشوار‌تر از شناخت
فعل‌های ساده است و به همین خاطر فراهم آوردن فهرستی از فعل‌های زبان (که شامل
فعل‌های مرکب نیز می‌شود) به همراه ساخت‌های ظرفیتی افعال، کمکی شایان برای کارهای
پردازشی است. از سوی دیگر، بر اساس نظریه وابستگی، ساخت بنیادین جمله را می‌توان
از
روی ساخت ظرفیتی فعل جمله به دست آورد و به همین دلیل بر اهمیت دانستن ساخت‌های
ظرفیتی فعل در متن‌های زبانی افزوده می‌شود.

"""


from collections import namedtuple
from pathlib import Path
from typing import Iterator

Verb = namedtuple(
    "Verb",
    (
        "past_light_verb",
        "present_light_verb",
        "prefix",
        "nonverbal_element",
        "preposition",
        "valency",
    ),
)


class VerbValencyReader:
    """این کلاس شامل توابعی برای خواندن پیکرهٔ ظرفیت نحوی افعال فارسی است.

    Args:
        valency_file: مسیر فایلِ پیکره.

    """

    def __init__(
        self: "VerbValencyReader", valency_file: str = "valency.txt",
    ) -> None:
        self._valency_file = valency_file

    def verbs(self: "VerbValencyReader") -> Iterator[Verb]:
        """افعال پیکره را برمی‌گرداند.

        Yields:
            فعل بعدی.

        """
        with Path.open(self._valency_file, encoding="utf-8") as valency_file:
            for line in valency_file:
                if "بن ماضی" in line:
                    continue

                line = line.strip().replace("-\t", "\t")
                parts = line.split("\t")
                if len(parts) == 6:
                    yield Verb(*parts)
